#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 

import logging
from socket import gethostname
from xiv_xsf.console import print_in_color
from xiv_xsf.scsy import exceptions as scsy_exceptions
from xiv_xsf.host.storage.devices import XIVStorageArrays
from hostattachment import definitions, wizard_assistant, translation, interactive, admin, exceptions
from hostattachment.exceptions import HostAttachmentError
from manager import XIVSystemManager
from hostattachment.definitions import ATTACHMENT_FC, ATTACHMENT_ISCSI

COLORS = ['cyan','dark_cyan']

@admin.try_and_catch
def define_xiv_storage_arrays(attachment):
    arrays = dict()
    arrays = get_xiv_storage_arrays()
    wizard_assistant.print_separator()
    if len(arrays.keys()) == 0:
        interactive.enter_to_proceed(translation.STORAGE_ARRAY_NOT_FOUND)
        return
    xiv_managers = list()
    for serial, devices in arrays.items():
        xiv_managers.append(XIVSystemManager(attachment, serial, devices))
    print_storage_arrays(xiv_managers)
    define_host(xiv_managers,attachment)
    interactive.enter_to_proceed("")

def get_xiv_storage_arrays():
    return XIVStorageArrays.get_all_xiv_storage_arrays()

def _get_arrays_properties(xiv_managers):
    arrays = []
    for machine in xiv_managers:
        array = {}
        value = machine.machine_serial
        array['serial'] = machine.machine_serial.replace('\t','').strip()
        array['version'] = machine.get_version_from_inquiry_data().replace('\t', '').strip()
        array['protocol'] = machine.get_protocols()
        if machine.is_host_defined():
            array['host_name'] = machine.hostname
            all_ports_defined = machine.is_host_defined_on_all_ports()
            if all_ports_defined:
                array['all_ports'] = 'All'
            else:
                array['all_ports'] = 'Not all'
            array['defined'] = True
        else:
            array['defined'] = False
            array['host_name'] = '--'
            array['all_ports'] = 'None'
        arrays.append(array)
    return arrays

def _get_formatted_array_properties(array):
    serial = array['serial'].ljust(10)
    version = array['version'].ljust(7)
    all_ports = array['all_ports'].ljust(15)
    protocol = array['protocol']
    protocol = ','.join(protocol)
    if not len(protocol):
        protocol = '--'
    protocol = protocol.ljust(10)
    hostname = array['host_name'].ljust(18)
    return serial, version, all_ports, protocol, hostname
    
def _print_arrays(arrays, colors, print_defined=True, print_undefined=True):
    for array in arrays:
        if array['defined']:
            if print_defined:
                defined = 'Yes'.ljust(14)
            else:
                continue
        else:
            if print_undefined:
                defined = 'No'.ljust(14)
            else:
                continue
        (serial, version, all_ports, protocol, hostname) = _get_formatted_array_properties(array)
        while True:
            try:
                print_in_color('%s%s%s%s%s%s' % (serial, version, defined, \
                all_ports, protocol, hostname), colors.next()[1])
                break
            except StopIteration:
                colors = enumerate(COLORS)
    return colors

def print_storage_arrays(xiv_managers, print_only_defined=False):
    print translation.STORAGE_ARRAY_FOUND
    print_in_color('%s%s%s%s%s%s' % (
                                'Serial'.ljust(10),
                                'Ver'.ljust(7),
                                'Host Defined'.ljust(14),
                                'Ports Defined'.ljust(15),
                                'Protocol'.ljust(10),
                                'Host Name(s)'.ljust(17)))
    colors = enumerate(COLORS)
    defined_arrays = []
    arrays = _get_arrays_properties(xiv_managers)
    colors = _print_arrays(arrays, colors, True, False)
    if print_only_defined:
        return None
    colors = _print_arrays(arrays, colors, False, True)

def define_host(xiv_managers, attachment):
    undefined_systems =[]
    for system in xiv_managers:
        if not system.is_host_defined_on_all_ports():
            protocol_string = 'FC' if attachment.protocol == ATTACHMENT_FC \
                            else 'iSCSI'
            if protocol_string in system.get_protocols() or len(system.get_protocols())==0:
                undefined_systems.append(system)
    if len(undefined_systems) == 0:
        if attachment.protocol == ATTACHMENT_FC:
            print translation.HOST_DEFINED_ON_ALL_ARRAYS_FC
        if attachment.protocol == ATTACHMENT_ISCSI:
            print translation.HOST_DEFINED_ON_ALL_ARRAYS_ISCSI
        return None
    if attachment.protocol == ATTACHMENT_FC:
        if not interactive.prompt(translation.ASK_TO_DEFINE_HOST_FC, "yes"):
            return None
    if attachment.protocol == ATTACHMENT_ISCSI:
        if not interactive.prompt(translation.ASK_TO_DEFINE_HOST_ISCSI, "yes"):
            return None
    hostname = interactive.prompt_hostname(gethostname())
    for system in undefined_systems:
        system.define_this_host(hostname)
    attachment.rescan()
    
@admin.try_and_catch
def admin_print_arrays(attachment):
    arrays = dict()
    arrays = get_xiv_storage_arrays()
    if len(arrays.keys()) == 0:
        print translation.STORAGE_ARRAY_NOT_FOUND
        return
    xiv_managers = list()
    for serial, devices in arrays.items():
        xiv_managers.append(XIVSystemManager(attachment, serial, devices))
    print_storage_arrays(xiv_managers)
    
def define_an_array(attachment, machine_serial, username, password, hostname):
    arrays = dict()
    arrays = get_xiv_storage_arrays()
    if len(arrays.keys()) == 0:
        print translation.STORAGE_ARRAY_NOT_FOUND
        return
    xiv_managers = list()
    for serial, devices in arrays.items():
        if serial == machine_serial:
            xiv_box = XIVSystemManager(attachment, serial, devices)
            if hostname is None:
                hostname = gethostname()
            xiv_box.define_this_host(hostname, username, password, False)
            return None
    logging.error('XIV system with serial %s not found' % machine_serial)
    raise HostAttachmentError(translation.STORAGE_ARRAY_REQUESTED_NOT_FOUND)
