#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 

import logging, sys, struct
from socket import gethostname

from xiv_xsf.scsy import exceptions as scsy_exceptions
from xiv_xsf.scsy.device import Device
from xiv_xsf.nextra.phase10.client import XCLIClient
from xiv_xsf.nextra.phase10.client import exceptions as p10client_exceptions
from xiv_xsf.nextra.phase10.client.xcli import exceptions as xcli_exceptions
from xiv_xsf.console import print_in_color
from hostattachment import definitions, wizard_assistant, translation, interactive, admin, exceptions
from hostattachment.platform import stdlib_platform
from hostattachment.exceptions import HostAttachmentError

from wrappers import scsy_exceptions_catcher, endpoints_exhausted

SCSY_MAX_BUFFER_SIZE = 2**17

class XIVSystemManager(object):
    def __init__(self, attachment, serial, devices):
        self.attachment = attachment
        self.devices = devices
        self.machine_serial = serial
        self.p10client = None
        self.hostname = None
    
    @scsy_exceptions_catcher
    def _get_inquiry_from_a_specific_device(self, device):
        scsy_device = Device()
        if ';' in device:
            split = device.split(';')
            scsy_device.open(split[0], split[1])
        else:
            scsy_device.open(device)
        inquiry = scsy_device.inquiry()
        scsy_device.close()
        return inquiry
    
    @scsy_exceptions_catcher    
    def _get_page_inquiry_from_a_specific_device(self, device, page):
        scsy_device = Device()
        if ';' in device:
            split = device.split(';')
            scsy_device.open(split[0], split[1])
        else:
            scsy_device.open(device)
        page00 = scsy_device.inquiry.evpd00()
        inquiry = None
        if int(str(page),16) in page00:
            page_inquiry = getattr(scsy_device.inquiry, 'evpd%s' % page)
            inquiry = page_inquiry()
        scsy_device.close()
        return inquiry
        
    def get_inquiry(self):
        inquiry = None
        for device in self.devices:
            try:
                inquiry = self._get_inquiry_from_a_specific_device(device)
            except Exception, e:
                logging.error('failed to inquiry device %s, error %s' % (device, str(e)))
                continue
        return inquiry
    
    def get_page_inquiry(self, page):
        inquiry = None
        for device in self.devices:
            try:
                inquiry = self._get_page_inquiry_from_a_specific_device(device, page)
            except Exception, e:
                logging.error('failed to page%s device %s, error %s' % (page, device, str(e)))
                continue
        return inquiry

    def get_page_inquiry_from_all_devices(self, page):
        inquiries = []
        for device in self.devices:
            try:
                inquiry = self._get_page_inquiry_from_a_specific_device(device, page)
                if inquiry is not None:
                    inquiries.append(inquiry)
            except Exception, e:
                logging.error('failed to page%s device %s, error %s' % (page, device, str(e)))
                continue
        return inquiries
    
    @endpoints_exhausted
    def _get_p10client_from_a_specific_device(self, username, password, device, ask_for_credentials):
        try:
            logging.debug('attempting to initiate an xcli session over scsi on device %s' % device)
            if ';' in device:
                split = device.split(';')
                raise HostAttachmentError("cannot define host when a volume is mapped to lun 0")
            else:
                return XCLIClient.connect_scsi(username, password, device, buffer_size=SCSY_MAX_BUFFER_SIZE)
        except (p10client_exceptions.IncorrectCredentials,
                p10client_exceptions.IncorrectPassword,
                p10client_exceptions.IncorrectUsername,
                p10client_exceptions.UnparseableErrorCodeDuringInstantiation):
            if ask_for_credentials:
                print_in_color(translation.STORAGE_INVALID_CREDENTIALS, 'red')
                return self._get_p10client(True, username, password)
            else:
                raise HostAttachmentError(translation.STORAGE_INVALID_CREDENTIALS)
        except (p10client_exceptions.IncorrectCredentials,\
                p10client_exceptions.IncorrectPassword,\
                p10client_exceptions.IncorrectUsername), e:
            raise HostAttachmentError(translation.STORAGE_INVALID_CREDENTIALS)
        except p10client_exceptions.UnparseableErrorCodeDuringInstantiation, e:
            if 'Incorrect password' in str(e) or 'User name does not exist' in str(e):
                raise HostAttachmentError(translation.STORAGE_INVALID_CREDENTIALS)
            else:
                raise HostAttachmentError("XCLIClient.connect_scsi returned unknown error: %s" % str(e))
        except Exception, e:
            raise HostAttachmentError("XCLIClient.connect_scsi returned unknown error: %s" % str(e))
        
    def _get_p10client(self, ask_for_credentials, username=None, password=None):
        if ask_for_credentials:
            (username,password) = interactive.prompt_username_and_password(self.machine_serial)
        logging.info(translation.STORAGE_ARRAY_CONNECTION)
        for device in self.devices:
            p10client = self._get_p10client_from_a_specific_device(username, password, device, ask_for_credentials)
            if p10client is not None:
                return p10client
        return None

    def get_version_from_inquiry_data(self):
        inquiry = self.get_inquiry()
        if inquiry is None:
            logging.error('failed to inquiry these devices: %s' % str(self.devices))
            return None
        return inquiry.product_revision_level
    
    def is_host_defined(self):
        inquiries = self.get_page_inquiry_from_all_devices(83)
        self.hostname = None
        if len(inquiries) == 0:
            logging.error('failed to inquiry these devices: %s' % str(self.devices))
            raise HostAttachmentError("failed to inquiry XIV system %s" % self.machine_serial)
        logging.debug('number of returned inquiries for XIV system %s = %s' % (self.machine_serial, len(inquiries)))
        for inquiry in inquiries:
            for container in inquiry.page_data.designation_descriptor:
                if not hasattr(container.designator, 'vendor_specific_identifier'):
                    continue
                if container.designator.vendor_specific_identifier.strip() == 'host=':
                    continue
                if 'host=' in container.designator.vendor_specific_identifier.strip():
                    hostname=container.designator.vendor_specific_identifier.split('=')[1].strip()
                    logging.debug('this host has one path to XIV system %s on which he is defined as %s' % \
                                  (self.machine_serial, hostname))
                    if self.hostname is None:
                        self.hostname = hostname
                    if hostname not in self.hostname.split(','):
                        self.hostname+=',%s' % hostname
        if self.hostname is not None:
            return True
        return False
    
    def is_host_defined_on_all_ports(self):
        inquiries = self.get_page_inquiry_from_all_devices(83)
        if len(inquiries) == 0:
            logging.error('failed to inquiry these devices: %s' % str(self.devices))
            raise HostAttachmentError("failed to inquiry XIV system %s" % self.machine_serial)
        logging.debug('number of returned inquiries for XIV system %s = %s' % (self.machine_serial, len(inquiries)))
        for inquiry in inquiries:
            for container in inquiry.page_data.designation_descriptor:
                if not hasattr(container.designator, 'vendor_specific_identifier'):
                    continue
                if container.designator.vendor_specific_identifier.strip() == 'host=':
                    logging.debug('got an inquiry on which the host is not defined: %s' % container)
                    return False
                if 'host=' in container.designator.vendor_specific_identifier.strip():
                    pass
        return True

    def get_protocols(self):
        connectivity_types = []
        inquiries = self.get_page_inquiry_from_all_devices(83)
        if len(inquiries) == 0:
            logging.error('failed to inquiry these devices: %s' % str(self.devices))
            raise HostAttachmentError("failed to inquiry XIV system %s" % self.machine_serial)
        logging.debug('number of returned inquiries for XIV system %s = %s' % (self.machine_serial, len(inquiries)))
        for inquiry in inquiries:
            for container in inquiry.page_data.designation_descriptor:
                if not hasattr(container, 'designator_type'):
                    continue
                if not container.designator_type == 'relative_target_port_identifier':
                    continue
                try:
                    i = struct.unpack(">H", container.designator[-2:])[0]
                    iscsi_port = (i & 0xFF) >= 0x20
                    interface_type = 'iSCSI' if iscsi_port else 'FC'
                    if interface_type not in connectivity_types:
                        connectivity_types.append(interface_type)
                except:
                    pass
                break
        connectivity_types.sort()
        return connectivity_types
    
    def _set_p10client(self, ask_for_credentials=True, username=None, password=None):
        if self.p10client is None:
            self.p10client = self._get_p10client(ask_for_credentials, username, password)
            return True
        if self.p10client is None:
            return False
        return True
        
    def define_this_host(self, hostname, username=None, password=None, interactive=True):
        if self.is_host_defined():
            if hostname not in self.hostname.split(','):
                logging.error('This host is already defined as host %s on XIV system %s' % (self.hostname, self.machine_serial))
                raise HostAttachmentError('This host is already defined as host %s on XIV system %s' % (self.hostname, self.machine_serial))
            if not self._set_p10client(interactive, username, password):
                return None
        else:
            if not self._set_p10client(interactive, username, password):
                return None
            self.try_host_define(hostname, interactive)
        self.host_add_self_ports()
        if hasattr(self.p10client, 'metadata_set'):
            self._set_platform_in_metadata()
            self._set_hak_version_in_metadata()
   
    @endpoints_exhausted
    def _p10client_host_define(self, hostname):
        return self.p10client.host_define(host=hostname)
    
    @endpoints_exhausted
    def _p10client_metadata_list(self):
        return self.p10client.metadata_list()
   
    @endpoints_exhausted
    def _host_add_port_fc(self, fcaddress):
        self.p10client.host_add_port(host=self.hostname, fcaddress=fcaddress)
    
    @endpoints_exhausted
    def _host_add_port_iscsi(self, iscsi_name):
        self.p10client.host_add_port(host=self.hostname, iscsi_name=iscsi_name)
  
    @endpoints_exhausted
    def _set_platform_in_metadata(self):
        self.p10client.metadata_set(name=self.hostname, object_type='host', \
                                   key='platform', value=stdlib_platform.platform())
    
    @endpoints_exhausted
    def _set_hak_version_in_metadata(self):
        self.p10client.metadata_set(name=self.hostname, object_type='host',
                                   key='hak_version', value=definitions.HOSTATTACHMENT_VERSION)
    @endpoints_exhausted
    def try_host_define(self, hostname, is_interactive):
        try:
            self._p10client_host_define(hostname)
        except xcli_exceptions.SoftExecutionFailedError, e:
            if 'Host name already exists' in str(e):
                if not is_interactive:
                    message = translation.STORAGE_ARRAY_HOST_ALREADY_EXISTS
                    logging.error(message)
                    raise exceptions.HostAttachmentError(str(e))
                answer = interactive.prompt(translation.STORAGE_ARRAY_HOST_ALREADY_EXISTS)
                if answer:
                    self.hostname = hostname
                else:
                    hostname = interactive.prompt_hostname(gethostname())
                    return self.try_host_define(hostname, True)
            else:
                raise e
        self.hostname = hostname
    
    def host_add_self_ports(self):
        for port in self.attachment.get_storage_ports():
            port_address = None
            port_type = None
            if self.attachment.protocol == definitions.ATTACHMENT_FC:
                port_address = port.wwpn
                port_type = definitions.ATTACHMENT_FC
            if self.attachment.protocol == definitions.ATTACHMENT_ISCSI:
                port_address = port
                port_type = definitions.ATTACHMENT_ISCSI
            self.host_add_self_port(port_address, port_type)
    
    def host_add_self_port(self, port_address, port_type):
        try:
            if port_type == definitions.ATTACHMENT_FC:
                self._host_add_port_fc(port_address) 
            if port_type == definitions.ATTACHMENT_ISCSI:
                self._host_add_port_iscsi(port_address)
        except xcli_exceptions.SoftExecutionFailedError, e:
            if 'Host with this port ID already defined' in str(e):
                logging.debug('a host with port ID %s is already defined on XIV system %s' % (port_address, self.machine_serial))
                pass
            else:
                raise e
                
    def host_remove_misdefined_port(self, port_address, port_type):
        for item in self.p10client.host_list():
            if port_type == definitions.ATTACHMENT_FC:
                if port_address in item.fc_ports.split(','):
                    self.p10client.host_remove_port(host=item.name, fcaddress=port_address)
                    break
            if port_type == definitions.ATTACHMENT_ISCSI:
                if port_address in item.iscsi_ports.split(','):
                    self.p10client.host_remove_port(host=item.name, iscsi_name=port_address)
                    break
 
    def get_key_from_metadata(self, name, object_type, key):
        try:
            metadata = self._p10client_metadata_list()
        except:
            raise NameError('metadata is not available')
        for info in metadata:
            if (name, object_type, key) == (info.name, info.object_type, info.key):
                return info.value
        raise KeyError('key %s for object name %s of type %s is not listed in metadata' % \
                       (key, name, object_type))
        
