#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import os
import logging

from hostattachment.exceptions import HostAttachmentError, RebootRequiredException
from hostattachment.definitions import ATTACHMENT_FC, ATTACHMENT_ISCSI
from hostattachment import translation

from xiv_xsf.console import print_in_color

CONFIGURATION_NAME_BACKUP_PATH = "backup_path"

class Attachment(object):
    def __init__(self, protocol, verbose_report=False):
        """
        Constructor.
        
        stop_config_on_reboot_exception enforces the configuration to stop after
        a configuration of a specific task has raised an exception.
        If it is set to False, when calling the config function of Attachment it
        will config all the tasks and raise a RebootRequiredException if one of
        the tasks raised it.
        The default value of stop_config_on_reboot_exception is True.

        stop_verify_on_error_exception enforces the configuration to stop verifying
        the rest of the tasks if one task raised a HostAttachmentError exception.
        If it is set to False, the verify function of Attachment will verify each
        of the appended tasks, and report all the errors to the end user.
        The default value of stop_verify_on_error_exception is True.
        
        By default, Attachment prints only errors in verify/config.
        verbose_report tells Attachment to also print the tasks the have been
        verified/configured successfully.
        xiv_attach will print only errors, and xiv_{fc/iscsi}_attach will
        print both errors and successes
        
        """
        self.protocol = protocol
        self.stop_config_on_reboot_exception = True
        self.stop_verify_on_error_exception  = True
        self.stop_verify_on_reboot_exception = True
        self.verbose_report = verbose_report
        self.tasks = list()

    def config(self):
        """
        Iterates over the list of tasks (self.tasks) and attempts to config it
        If stop_config_on_reboot_exception is set, config will stop when a
        reboot exception has been raised. If it is not set, it will continue to
        config each task, and raise the reboot exception at the end.
        """
        reboot_required = False
        for task in self.tasks:
            if task.protocol is None or self.protocol == task.protocol:
                try: 
                    self.try_config_task(task)
                except RebootRequiredException, e:
                    if self.stop_config_on_reboot_exception:
                        raise RebootRequiredException()
                    else:
                        logging.debug("RebootRequiredException suppressed: %s" % str(e))
                        reboot_required = True
        if reboot_required:
            raise RebootRequiredException()
    
    def unconfig(self):
        """
        Iterated over the list of tasks and attemps to unconfig it
        If stop_config_on_reboot_exception is set, unconfig will stop when a 
        reboot exception has been raised. If it is not set, it will continue 
        to config each task and raise the reboot exception at the end"
        """
        reboot_required = False
        for task in self.tasks:
            if task.protocol is None or self.protocol == task.protocol:
                try:
                    self.try_unconfig_task(task)
                except RebootRequiredException, e:
                    if self.stop_config_on_reboot_exception:
                        raise RebootRequiredException()
                    else:
                        logging.debug("RebootRequiredException suppressed: %s" % str(e))
                        reboot_required = True
        if reboot_required:
            raise RebootRequiredException()
    
    def verify(self):
        """
        Iterates over the list of tasks (self.tasks) and attempts to verify
        each of the tasks
        If no task returned False, verify returns True
        If at least one task return False, verify retuns False
        If an exception was raised and attachment is configured to stop on it,
        it will raise the exception and not continue to the next task.
        Else, it willl continue to the next task and throw the exception at the end.
        """
        error_exception = False
        most_recent_error_exception = None
        reboot_exception = False
        result = True
        for task in self.tasks:
            if task.protocol is None or self.protocol == task.protocol:
                try:
                    if not self.try_verify_task(task):
                        result = False
                except HostAttachmentError, e:
                    if self.stop_verify_on_error_exception:
                        raise HostAttachmentError(e)
                    else:
                        logging.debug("HostAttachmentError suppressed: %s" % str(e))
                        error_exception = True
                        most_recent_error_exception = e
                except RebootRequiredException, e:
                    if self.stop_verify_on_reboot_exception:
                        raise RebootRequiredException()
                    else:
                        logging.debug("RebootRequiredException suppressed: %s" % str(e))
                        reboot_exception = True
        if error_exception:
            raise HostAttachmentError(most_recent_error_exception)
        if reboot_exception:
            if result == True:
                raise RebootRequiredException()
        if result is True and self.verbose_report:
            print translation.HOST_ALREADY_CONFIGURED
        return result
    
    def try_verify_task(self, task):
        """
        Attempts to verify a given task.
        If verbose_report is set, the task name will be printed, 
        with "OK" for success, "NOT OK" for false, and ERROR or REBOOT in case of exceptions.
        """
        logging.debug('Task: verifying %s' % task.name)
        result = False
        error_exception = None
        if self.verbose_report:
            self.print_task_name(task)
        try:
            if task.verify():
                if self.verbose_report:
                    print_in_color("    OK", 'dark_green')
                return True
            else:
                if self.verbose_report:
                    print_in_color("NOT OK", 'yellow')
                return False
        except HostAttachmentError, e:
            if self.verbose_report:
                print_in_color(" ERROR", 'red')
                formatted_text = "    Reason: %s" % str(e).splitlines()[0]
                logging.error(formatted_text)
                print_in_color(formatted_text.ljust(80), 'dark_red')
            raise HostAttachmentError(e)
        except RebootRequiredException, e:
            if self.verbose_report:
                print_in_color(" REBOOT", 'cyan')
            raise RebootRequiredException
        return None
              
    def try_config_task(self, task):
        """
        Attempts to config a given task.
        If verbose_report is set, a message will be printed if the verify
        of the task failed or if it succeeded.
        If it is not set, a message will be printed only if an exceptation was
        raised.
        """
        if self.verbose_report:
            self.print_task_name(task)
        try:
            if not task.verify():
                logging.debug('Task: configuring %s' % task.name)
                task.configure()
            if self.verbose_report:
                print_in_color ("    OK", 'dark_green')
        except HostAttachmentError, e:
            if self.verbose_report:
                print_in_color (" ERROR", 'red')
                formatted_text = "    Reason: %s" % str(e).splitlines()[0]
                logging.error(formatted_text)
                print_in_color(formatted_text.ljust(80), 'dark_red')
            raise HostAttachmentError(e)
        except RebootRequiredException:
            if self.verbose_report:
                print_in_color("REBOOT", 'cyan')
            raise RebootRequiredException()
    
    def try_unconfig_task(self, task):
        """
        Attempts to unconfig a given task.
        If verbose_report is set, a message will be printed if the verify
        of the task failed or if it succeeded.
        If it is not set, a message will be printed only if an exceptation was
        raised.
        """
        if self.verbose_report:
            self.print_task_name(task)
        try:
            if task.verify():
                if hasattr(task, 'unconfigure'):
                    logging.debug('Task: unconfiguring %s' % task.name)
                    task.unconfigure()
            if self.verbose_report:
                print_in_color ("    OK", 'dark_green')
        except HostAttachmentError, e:
            if self.verbose_report:
                print_in_color (" ERROR", 'red')
                formatted_text = "    Reason: %s" % str(e).splitlines()[0]
                logging.error(formatted_text)
                print_in_color(formatted_text.ljust(80), 'dark_red')
            raise HostAttachmentError(e)
        except RebootRequiredException:
            if self.verbose_report:
                print_in_color("REBOOT", 'cyan')
            raise RebootRequiredException()
        except NotImplementedError, e:
            if self.verbose_report:
                print_in_color ("  SKIP", 'yellow')
    
    def print_task_name(self, task):
        """
        prints the task name
        """
        formatted_text = "%s..." % task.name
        logging.info(formatted_text)
        print_in_color (formatted_text.ljust(73), newline=False)

    def set_additional_options(self, parser):
        """
        Set any platform-specific options for the xiv_*_admin cli tools
        This method is called uppon the generation of the cli options
        you may add parser options and modifications here
        """
        pass

    def handle_additional_options(self, options=None, is_interactive=False):
        """
        Process any additional options added to the parser by 
        this attachment. If options is None, an interactive (xiv_attach)
        mode is assumed, and you should read the options from the user's
        input
        """
        pass

class FCPort(object):
    def __init__ (self, wwpn, product, model, device_filename=None):
        self.wwpn = wwpn
        self.product = product
        self.model = model
        self.device_filename = device_filename

    def __str__(self):
        fc_port_string = self.wwpn + ": "
        if self.device_filename:
            fc_port_string += self.device_filename + ": "
        fc_port_string += "[%s]: %s" % (self.product, self.model)
        return fc_port_string

class Task(object):
    """
    Task represents a task to be performed on the host during the attachment procedure.
    A task can be used in FC-attachment-only, iSCSI-attachments-only, or both.
    Each task must contain the following properties:
    name        The name of the task, which will be printed to the standard output
    protocol    The protocol defines in which attachments the HAK will perform this task.
                For FC-only attachments, its value needs to be ATTACHMENT_FC
                For iSCSI-only attachments, its value needs to be ATTTACHMENT_ISCSI
                For both types, its value needs to be None
    """
    def __init__(self, name, protocol=None):
        """
        The constructor may raise a HostAttachmentError exception if a critical error
        has occurred, which forces the HAK to abort
        """
        self.name = name
        if protocol is not None:
            if protocol != ATTACHMENT_FC and protocol != ATTACHMENT_ISCSI:
                raise HostAttachmentError("invalid protcol")
        self.protocol = protocol
    
    def verify(self):
        """
        Verifies if the task has been configured,
        or if it is valid.
        If the task has been verified, the function returns True
        If the verification has failed due to serious error, the function raises a HostAttachmentError
        If the task is just not configured, the function returns False with a
        message to report to the standard output
        """
        raise NotImplementedError
    
    def configure(self):
        """
        Perform the configuration part of the task.
        If the configuation has passed successfully, the functions returns None
        If the configuration, the function raises a HostAttachmentError
        with the reason for the failure
        If the task requires a reboot, 
        the functions raises a RebootRequiredException
        """
        raise NotImplementedError

    def unconfigure(self):
        """
        Performs the opposite of the configure.
        This method is to be used when the end user want to uninstall host attachment kit
        and "detach" the host from xiv
        Implementing this function is optional, as the task can be about something generic
        that should not be removed upon removal of hostattachment kit
        """
        raise NotImplementedError

class AttachmentConfiguration(dict):
    singleton = None

    def __init__(self):
        self.config = dict()

    def set(self, config_name, value):
        self[config_name] = value
    
    @property
    def backup_path(self):
        return self.get(CONFIGURATION_NAME_BACKUP_PATH)

    def set_backup_path(self, backup_path):
        self.set(CONFIGURATION_NAME_BACKUP_PATH, backup_path)
        
    @classmethod
    def getConfiguration(cls):
        if cls.singleton is None:
            cls.singleton = AttachmentConfiguration()
        return cls.singleton 




