#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import sys
from getpass import getpass
import translation, definitions

from xiv_xsf.console import print_in_color
from iscsicommon import test_iscsi_connectivity

ISCSI_PORT = 3260
DEFAULT_VALID_REPLIES = {'yes': True, 'y': True, 'no': False, 'n': False}

class NoReply(object):
    pass

def _prompt_print_message(message, default, default_color):
    print '%s' % message,
    if default:
        print '[default: ',
        print_in_color(default, default_color, newline=False)
        print ']:',

def _prompt_lower_reply_and_valid_replies(reply, valid_replies):
    replies = None
    if type(valid_replies) is dict:
        replies = dict()
        for (key, value) in valid_replies.iteritems():
            replies[key.lower()]=value
    if reply is None:
        return None, replies
    return reply.lower(), replies

def _prompt_get_reply(default_answer):
    reply = raw_input() or default_answer
    try:
        reply = reply.strip()
    except AttributeError:
            pass
    return reply

def _prompt_check_reply(reply, valid_replies):
    if type(valid_replies) is dict:
        if reply not in valid_replies.keys():
            print_in_color('%s ' % translation.INPUT_INVALID_PREFIX,
                           'red', newline=False)
            print ', '.join(valid_replies.keys())
        else:
            return valid_replies[reply]
    else:
        if reply is None:
            return NoReply()
        if len(reply):
            return reply
        else:
            print_in_color('%s ' % translation.INPUT_INVALID_PREFIX_NO_VALID_REPLIES,
                           'red', newline=False)
    return NoReply()

def prompt(message, default_answer="no", valid_replies=DEFAULT_VALID_REPLIES, 
           case_sensitive=False, default_color='green'):
    try:
        while True:
            _prompt_print_message(message, default_answer, default_color)
            reply = _prompt_get_reply(default_answer)
            if not case_sensitive:
                reply, valid_replies = \
                     _prompt_lower_reply_and_valid_replies(reply, valid_replies)
            reply = _prompt_check_reply(reply, valid_replies)
            if type(reply) is NoReply:
                continue
            return reply
    except (KeyboardInterrupt, EOFError), e:
        print ""
        raise e

def enter_to_proceed(message):
    print message + "\nPress ",
    print_in_color('[ENTER]', 'green', newline=False)
    print "to proceed. ",
    try:
        nothing = getpass("")
    except (KeyboardInterrupt, EOFError), e:
        print ''
        raise e
    print ''
    
def enter_to_exit(message):
    print message + "\nPress ",
    print_in_color('[ENTER]', 'green', newline=False)
    print "to exit. ",
    try:
        nothing = getpass("")
    except (KeyboardInterrupt, EOFError), e:
        pass
    print ''

def prompt_iscsi_address(message):
    return prompt('%s ' % message, None, None, False)

def prompt_iscsi_chap_name():
    try:
        message = translation.ISCSI_CHAP_NAME + " "
        while True:
            reply = raw_input(message).strip()
            if len(reply)>0:
                return reply
            print translation.INPUT_INVALID_PREFIX, \
                  translation.ISCSI_CHAP_NAME_INVALID
            continue
    except (KeyboardInterrupt, EOFError), e:
        print ""
        raise e

def prompt_iscsi_chap_secret():
    try:
        message = translation.ISCSI_CHAP_SECRET + " "
        while True:
            reply = raw_input(message).strip()
            if len(reply)>=12 and len(reply)<=16:
                return reply
            print translation.INPUT_INVALID_PREFIX, \
                  translation.ISCSI_CHAP_SECRET_INVALID
            continue
    except (KeyboardInterrupt, EOFError), e:
        print ""
        raise e

def prompt_hostname(default):
    return prompt(translation.STORAGE_ARRAY_HOSTNAME, default, None, True)
    
def prompt_username_and_password(serial):
    username = 'admin'
    password = ''
    try:
        print translation.STORAGE_ARRAY_USERNAME + " ",
        print_in_color(serial, 'cyan', newline=False)
        print ': [default: ',
        print_in_color('admin', 'green', newline=False)
        print ']: ',
        reply = raw_input().strip()
        if len(reply) > 0:
            username = reply
        print translation.STORAGE_ARRAY_PASSWORD_PART_1 + " ",
        print_in_color(username, 'green', newline=False)
        print translation.STORAGE_ARRAY_PASSWORD_PART_2 + " ",
        print_in_color(serial, 'cyan', newline=False)
        sys.stdout.flush()
        reply = getpass(': ')
        print ''
        password = reply
    except (KeyboardInterrupt, EOFError), e:
        print ""
        raise e
    return username, password
        

def choose_connectivity_type():
    default = None
    valid_replies={
        "iscsi": definitions.ATTACHMENT_ISCSI, "i": definitions.ATTACHMENT_ISCSI,
        "fc": definitions.ATTACHMENT_FC, "f": definitions.ATTACHMENT_FC,
        "abort": None
    }
    try:
        while True:
            print '%s, ' % translation.ATTACHMENT_TYPE_PROMPT, 
            print_in_color('[f]c / [i]scsi', 'green', newline=False)
            print ':',
            reply = _prompt_get_reply(None)
            reply, valid_replies = _prompt_lower_reply_and_valid_replies(reply, valid_replies)
            reply = _prompt_check_reply(reply, valid_replies)
            if type(reply) is NoReply:
                continue
            return reply
    except (KeyboardInterrupt, EOFError), e:
        print ""
        raise e
