#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 

import logging, os

import translation, definitions
from optparse import OptionParser, OptionGroup
from hostattachment.exceptions import HostAttachmentError, VerifyFailedError
from definitions import ATTACHMENT_FC, ATTACHMENT_ISCSI
from hostattachment import admin, array

DEFAULT_PARSER = OptionParser()
MAIN_TASK_OPTIONS = ["verify", "configure", "rescan", "connect", "disconnect",
                     "print_storage_ports", "version", "define", "list"]

def _verify_conflicting_options(options):
    """
    Makes sure the user doesn't call two commands at once, i.e.
    configure + print + verify. It might cause troubles (order-wise),
    so we'd better just disallow it"""
    found = False
    for task in MAIN_TASK_OPTIONS:
        try:
            if getattr(options, task):
                if (found):
                    raise HostAttachmentError(translation.CONFLICTING_COMMANDS)
                found = True
        except AttributeError, e:
            pass

def _verify_host_define_options(options):
    if not options.xcli_serial:
        raise HostAttachmentError(translation.MISSING_MACHINE_SERIAL)
    if not options.xcli_user:
        raise HostAttachmentError(translation.MISSING_MACHINE_USER)
    if not options.xcli_pass:
        raise HostAttachmentError(translation.MISSING_MACHINE_PASSWORD)

def _verify_iscsi_chap_options(options):
    if options.iscsi_name:
        if options.iscsi_secret:
            if len(options.iscsi_name) == 0:
                raise HostAttachmentError(translation.ISCSI_CHAP_NAME_INVALID)
            length = len(options.iscsi_secret)
            if length<12 or length>16:
                raise HostAttachmentError(translation.ISCSI_CHAP_SECRET_INVALID)
        else:
            raise HostAttachmentError(translation.ISCSI_CHAP_BOTH_NAME_AND_SECRET)
    else:
        if options.iscsi_secret:
            raise HostAttachmentError(translation.ISCSI_CHAP_BOTH_NAME_AND_SECRET)
            

def get_default_option_parser():
    parser = DEFAULT_PARSER
    parser.add_option("-v", "--version", action="store_true", dest="version", 
                      default=False,help="print hostattachment kit version")
    parser.add_option("-V", "--verify", action="store_true", dest="verify", 
                      default=False, help="Verify host configuration tasks")
    parser.add_option("-C", "--configure", action="store_true", dest="configure", 
                      default=False, help="Configure this host for attachment")
    parser.add_option("-R", "--rescan", action="store_true", dest="rescan", 
                      default=False, help="Rescan devices")
    parser.add_option("-D", "--define", action="store_true", dest="define",
                      default=False, help="Define this host on a system. See 'Host Definition Options'")
    parser.add_option("-L", "--list", action="store_true", dest="list",
                      default=False, help="List attached XIV systems")
    return get_define_option_parser(parser)


def get_attachment_option_parser(attachment):
    parser = None
    if attachment.protocol == ATTACHMENT_FC:
        parser = get_fc_option_parser()
    elif attachment.protocol == ATTACHMENT_ISCSI:
        parser = get_iscsi_option_parser()
    else:
        raise HostAttachmentError("Error recognizing protocol of attachment")
    attachment.set_additional_options(parser)
    return parser

def get_define_option_parser(parser):
    define_group = OptionGroup(parser, "Host Definition Options")
    define_group.add_option("-u", "--user", action="store", dest="xcli_user",
                            help="Username for XCLI", metavar="USERNAME")
    define_group.add_option("-p", "--pass", action="store", dest="xcli_pass",
                            help="Password for XCLI", metavar="PASSWORD")
    define_group.add_option("-H", "--hostname", action="store", dest="xcli_host", default=None,
                            help="Hostname for this host (optional). Unless specified, the os hostname will be used", metavar="HOSTNAME")
    define_group.add_option("-S", "--serial", action="store", dest="xcli_serial",
                            help="Serial of machine. See parameter '--list'.", metavar="SERIAL")
    parser.add_option_group(define_group)
    return parser

def get_fc_option_parser():
    parser = get_default_option_parser()
    parser.add_option("-P", "--print", action="store_true", 
                      dest="print_storage_ports", default=False,
                      help="Print WWPN of HBA devices")
    return parser

def get_iscsi_option_parser():
    parser = get_default_option_parser()
    parser.add_option("-P", "--print", action="store_true", dest="print_storage_ports",
                      default=False, help="Print iSCSI initiator name and alias")
    parser.add_option("-c", "--connect", action="store", dest="connect", 
                      help="Discover new iSCSI target", metavar="ISCSI_INTERFACE_IP")
    parser.add_option("-d", "--disconnect", action="store", dest="disconnect", 
                      help="Disconnect from an iSCSI target", metavar="ISCSI_INTERFACE_IP")
    iscsi_group = OptionGroup(parser, "iSCSI CHAP Options")
    iscsi_group.add_option("-n", "--name", action="store", dest="iscsi_name", 
                           help="CHAP name", metavar="ISCSI_CHAP_NAME", default=None)
    iscsi_group.add_option("-s", "--secret", action="store", dest="iscsi_secret", 
                           help="CHAP secret", metavar="ISCSI_CHAP_SECRET", default=None)
    parser.add_option_group(iscsi_group)
    return parser

def default_handler(attachment, options):
    _verify_conflicting_options(options)
    done = True
    attachment.handle_additional_options(options)
    if options.version:
        print definitions.HOSTATTACHMENT_VERSION
    elif options.verify:
        if not attachment.verify():
            raise VerifyFailedError()
    elif options.configure:
        attachment.config()
    elif options.print_storage_ports:
        for port in attachment.get_storage_ports():
            print str(port)
    elif options.rescan:
        attachment.rescan()
    elif options.list:
        array.admin_print_arrays(attachment)
    elif options.define:
        _verify_host_define_options(options)
        array.define_an_array(attachment, options.xcli_serial, options.xcli_user,
                              options.xcli_pass, options.xcli_host)
    else:
        done = False
    return done

def fc_handler(attachment, options):
    return default_handler(attachment, options)

def iscsi_handler(attachment, options):
    if default_handler(attachment, options):
        return True
    _verify_iscsi_chap_options(options)
    if options.connect:
        attachment.connect_iscsi(options.connect, options.iscsi_name, options.iscsi_secret)
        return True
    elif options.disconnect:
        attachment.disconnect_iscsi(options.disconnect)
        return True
    return False
