#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import platform
import re
import logging

from hostattachment.classes import Attachment
from hostattachment.exceptions import HostAttachmentError
from hostattachment.platform.linux import messages
from hostattachment.definitions import ATTACHMENT_FC, ATTACHMENT_ISCSI

from xiv_xsf.host.os.linux import get_distro_version
from xiv_xsf.host.os.linux.kernel_modules import load_kernel_module

class LinuxAttachment(Attachment):
    def __init__(self, protocol, verbose_report = False):
        Attachment.__init__(self, protocol, verbose_report)
        if not self.is_platform_supported():
            raise HostAttachmentError(messages.UNSUPPORTED_OS)

        self._load_needed_kernel_modules()
        self.tasks = get_tasks()

    def _load_needed_kernel_modules(self):
        # sg is needed for various HAK tasks
        load_kernel_module("sg")

        # m.p. modules are not loaded by default on some cases
        # e.g. RHEL4.6, right after installing multipath tools, before the first
        # boot
        load_kernel_module("dm-multipath")
        load_kernel_module("dm-round-robin")

    @staticmethod
    def is_platform_supported():
        if platform.system() != "Linux":
            raise HostAttachmentError(\
                "This host attachment version can be run on Linux operating system only.")

        (distro, major, minor) = get_distro_version()

        if distro == "redhat":
            if major == 4 and minor >= 5:
                return True # RHEL >= 4.5
            elif major == 5 and minor >= 1:
                return True # >= 5.1

        elif distro == "suse":
            if major == 10 and minor >= 2:
                return True # 10 >= sp2
            elif major == 11:
                return True # 11

        return False

    def is_platform_officially_supported(self):
        return self.is_platform_supported()

def get_os_string():
    """Returns os + version strings such as redhat4 or SuSE10"""
    dist = platform.dist()
    name = dist[0]

    # Get most significant number of the version
    match = re.compile("^(\d+)").search(dist[1])
    if match is None:
        raise HostAttachmentError("Unrecognized OS version string '%s'" % dist[1])

    version = match.groups()[0]

    ret_str = "%s%s" % (name, version)

    return ret_str


class ISCSIAttachment(LinuxAttachment):
    def __init__(self, verbose_report = False):
        from hostattachment.platform.linux.iscsi import get_iscsiadm_object,\
                                                  get_initiatorname_object, \
                                                  detect_iscsi_method
        LinuxAttachment.__init__(self, ATTACHMENT_ISCSI, verbose_report)

        # This will throw a special exception if iscsi is not detected
        detect_iscsi_method()

        self.iscsiadm = get_iscsiadm_object()
        self.initiatorname = get_initiatorname_object()

    def rescan(self):
        self.iscsiadm.iscsi_scan()

    def connect_iscsi(self, discovery_address, chap_name=None, chap_password=None):
        # TODO MAKE SURE ITS CONFIGURED
        self.iscsiadm.connect(discovery_address, chap_name, chap_password)

    def disconnect_iscsi(self, discovery_address):
        # TODO MAKE SURE ITS CONFIGURED
        self.iscsiadm.disconnect(discovery_address)

    def get_storage_ports(self):
        from hostattachment.platform.linux.iscsi.classes import \
                                                        BadInitiatorNameError
        # TODO handle no-iqn, no file or bad file
        try:
            iqn = self.initiatorname.get_iqn()
        except BadInitiatorNameError, e:
            raise HostAttachmentError(messages.BAD_INITIATOR_NAME)
        return [iqn]

class FCAttachment(LinuxAttachment):
    def __init__(self, verbose_report = False):
        LinuxAttachment.__init__(self, ATTACHMENT_FC, verbose_report)

    def get_storage_ports(self):
        from hostattachment.classes import FCPort
        from hostattachment.platform.linux.fc.fc_hosts import FiberChannelHosts
        fc_hosts = FiberChannelHosts()
        fc_ports = []
        scsi_hosts = fc_hosts.get_scsi_hosts()
        for scsi_host in scsi_hosts:
            fc_ports.append(FCPort(scsi_host.wwn, scsi_host.type, "N/A"))

        return fc_ports

    def rescan(self):
        from hostattachment.platform.linux.fc.fc_hosts import FiberChannelHosts
        fc_hosts = FiberChannelHosts()
        logging.info(messages.RESCANNING_MIGHT_TAKE_A_WHILE)
        fc_hosts.scan_all_hosts()
        logging.info(messages.RESCAN_ENDED)

def get_multipath_tasks():
    tasks = []
    from hostattachment.platform.linux.multipath.classes import \
            UdevRulesTask, MultipathConfTask, MultipathdServiceTask
    tasks.append(UdevRulesTask())
    tasks.append(MultipathConfTask())
    tasks.append(MultipathdServiceTask())

    return tasks

def get_iscsi_tasks():
    tasks = []

    from hostattachment.platform.linux.iscsi import IscsiInitiatorTask, \
                                                    IscsidConfTask, \
                                                    IscsiServiceTask
    tasks.append(IscsiInitiatorTask())
    tasks.append(IscsidConfTask())
    tasks.append(IscsiServiceTask())

    return tasks

def get_tasks():
    tasks = []
    tasks += get_multipath_tasks()
    tasks += get_iscsi_tasks()

    return tasks
