#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import logging
import glob
import re
import os

from hostattachment.exceptions import HostAttachmentError, \
                                      IscsiNotDetectedError
from hostattachment.classes import Task
from hostattachment.definitions import ATTACHMENT_ISCSI
from hostattachment.platform.linux import messages
from hostattachment.platform.linux.service import GenericServiceManager
from xiv_xsf.host.os.linux import get_distro_version
from hostattachment.platform.linux.scsi_device_manager import remove_bad_luns

OPEN_ISCSI = 0
OLD_LINUX_ISCSI = 1

def detect_iscsi_method():
    """Decides if it's OpenIscsi (modern) or LinuxIscsi (deprecated), and
    returns the right object

    Of all currently supported platforms, RHEL4 is the only LinuxIscsi one"""
    if os.path.exists("/sbin/iscsiadm"):
        logging.debug("OpenIscsi was detected")
        return OPEN_ISCSI
    elif os.path.exists("/sbin/iscsi-device"):
        logging.debug("LinuxIscsi was detected")
        return OLD_LINUX_ISCSI
    else:
        raise IscsiNotDetectedError(messages.NO_ISCSI_WAS_DETECTED)

def get_initiatorname_object():
    """Detect and return the appropiate InitiatorName object"""
    iscsi_method = detect_iscsi_method()

    if iscsi_method == OPEN_ISCSI:
        from hostattachment.platform.linux.iscsi.open_iscsi import \
                                                     InitiatorNameOpenIscsi
        return InitiatorNameOpenIscsi()
    else: # OLD_LINUX_ISCSI
        from hostattachment.platform.linux.iscsi.linux_iscsi import \
                                                     InitiatorNameLinuxIscsi
        return InitiatorNameLinuxIscsi()

def get_iscsid_object():
    """Detect and return the appropiate iscsid-conf object"""
    iscsi_method = detect_iscsi_method()

    if iscsi_method == OPEN_ISCSI:
        from hostattachment.platform.linux.iscsi.open_iscsi import \
                                                     IscsidConfOpenIscsi
        return IscsidConfOpenIscsi()
    else: # OLD_LINUX_ISCSI
        from hostattachment.platform.linux.iscsi.linux_iscsi import \
                                                     IscsidConfLinuxIscsi
        return IscsidConfLinuxIscsi()

def get_iscsi_service_object():
    iscsi_method = detect_iscsi_method()
    init_script_name = None

    if iscsi_method == OPEN_ISCSI:
        (distro, major, minor) = get_distro_version()
        if distro == "redhat":
            init_script_name = "iscsi"
        else: # suse. We wouldn't get here if it's an unsupported version
            init_script_name = "open-iscsi"
    else: # OLD_LINUX_ISCSI
        init_script_name = "iscsi"

    return GenericServiceManager(init_script = init_script_name,
                                 daemon_name = "iscsid")

def get_list_of_iscsi_hosts():
    iscsi_hosts = []

    for iscsi_host in glob.glob("/sys/class/iscsi_host/*"):
        match = re.compile("^host(\d+)").search(os.path.basename(iscsi_host))
        if (match is not None):
            iscsi_hosts.append(match.groups()[0])

    return iscsi_hosts

def remove_bad_luns_from_all_iscsi_hosts():
    iscsi_hosts = get_list_of_iscsi_hosts()
    for iscsi_host in iscsi_hosts:
        remove_bad_luns(iscsi_host)

def get_iscsiadm_object():
    iscsi_method = detect_iscsi_method()
    if iscsi_method == OPEN_ISCSI:
        from hostattachment.platform.linux.iscsi.open_iscsi.connection_manager\
                                                      import ConnectionManager
        return ConnectionManager()
    else: # OLD_LINUX_ISCSI
        from hostattachment.platform.linux.iscsi.linux_iscsi.connection_manager\
                                                       import ConnectionManager
        return ConnectionManager()
 
class IscsiInitiatorTask(Task):
    def __init__(self):
        Task.__init__(self, "iSCSI initiator configuration", ATTACHMENT_ISCSI)

    def configure(self):
        initiatorname = get_initiatorname_object()
        initiatorname.configure()

    def verify(self):
        initiatorname = get_initiatorname_object()
        return initiatorname.is_configured()

class IscsidConfTask(Task):
    def __init__(self):
        Task.__init__(self, "/etc/iscsid.conf file", ATTACHMENT_ISCSI)

    def configure(self):
        iscsid_conf = get_iscsid_object()
        iscsid_conf.configure()

    def verify(self):
        iscsid_conf = get_iscsid_object()
        return iscsid_conf.is_configured()

class IscsiServiceTask(Task):
    def __init__(self):
        Task.__init__(self, "iscsi service", ATTACHMENT_ISCSI)

    def configure(self):
        iscsi_service = get_iscsi_service_object()
        iscsi_service.configure(ignore_service_retval = True)

    def verify(self):
        iscsi_service = get_iscsi_service_object()
        return iscsi_service.is_configured()
