#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import logging
from xiv.dtypes import Bunch

from hostattachment.platform.linux import definitions as linux_hak_defs
from hostattachment.platform.linux.iscsi.classes import ConfigFileReader, \
    ConfigFileWriter, IIscsidConf, IConfigFile
from hostattachment.classes import AttachmentConfiguration

class IscsiTarget(Bunch):
    def __init__(self, ip_address, chap_name=None, chap_password=None):
        Bunch.__init__(self)
        self.ip_address = ip_address
        self.chap_name = chap_name
        self.chap_password = chap_password
    
class ConfigFileWriterLinuxIscsi(ConfigFileWriter):
    """iscsid.conf configuration handler (Old Linux-iSCSI)"""
    def __init__(self, filepath, settings = {}, targets=[]):
        ConfigFileWriter.__init__(self, filepath, settings)
        self.targets = targets

    def append_settings(self):
        for (name, value) in self.settings.iteritems():
            # NOTE: Unlike open_iscsi we don't use spaces here between
            # key & value, there might be a good reason for it
            self.lines.append('%s=%s' % (name, value))
        self.append_targets()

    def append_targets(self):
        for target in self.targets:
            self.lines.append('DiscoveryAddress=%s' % target.ip_address)
            if target.chap_name is not None:
                self.lines.append('    OutgoingUsername=%s' % target.chap_name)
            if target.chap_password is not None:
                self.lines.append('    OutgoingPassword=%s' % target.chap_password)

class ConfigFileLinuxIscsi(IConfigFile):
    def __init__(self, filepath):
        self.targets = []
        self.filepath = filepath
        attachment_configuration = AttachmentConfiguration.getConfiguration()
        self.backup_path = attachment_configuration.backup_path
        self.settings = {}
        self.reader = ConfigFileReader(self.filepath)
        self.read()
    
    def get_item_from_specific_config_line(self, line_number):
        if line_number < 0:
            return None, None
        if line_number >= len(self.reader.config_lines):
            return None, None
        line = self.reader.config_lines[line_number]
        (name, value) = self.split_n_strip_line(line)
        logging.debug('found %s=%s' % (name, value))
        return name, value

    def analyze(self):
        for index in range (0, len(self.reader.config_lines)):
            (name, value) = self.get_item_from_specific_config_line(index)
            if name == 'DiscoveryAddress':
                chap_name = None
                chap_password = None
                (key_2, value_2) = self.get_item_from_specific_config_line(index+1)
                if key_2 == 'OutgoingUsername' or key_2 == 'Username':
                    chap_name = value_2
                (key_2, value_2) = self.get_item_from_specific_config_line(index+2)
                if key_2 == 'OutgoingPassword' or  key_2 == 'Password':
                    chap_password = value_2
                self.add_target(value, chap_name, chap_password)
            elif name in ['OutgoingPassword', 'Password', 'OutgoingUsername', 'Username']:
                continue
            else:
                self.add_setting(name, value)

    def add_target(self, ip_address, chap_name, chap_password):
        for target in self.targets:
            if target.ip_address == ip_address:
                return None
        self.targets.append(IscsiTarget(ip_address, chap_name, chap_password))

    def remove_target(self, ip_address):
        for target in self.targets:
            if target.ip_address == ip_address:
                self.targets.remove(target)
        
    def write_to_file(self, filepath=''):
        if filepath == '':
            filepath = self.filepath 
        logging.debug('settings = %s' % self.settings)
        logging.debug('targets = %s' % self.targets)
        writer = ConfigFileWriterLinuxIscsi(filepath, self.settings, \
                                            self.targets)
        return writer.write_to_file()

    def is_empty(self):
        if len(self.settings) == 0 and len(self.targets) == 0:
            return True
        return False
        
class IscsidConfLinuxIscsi(IIscsidConf):
    def __init__(self):
        attachment_configuration = AttachmentConfiguration.getConfiguration()
        self.backup_path = attachment_configuration.backup_path
        self.source_file = linux_hak_defs.ISCSI_CONF_SOURCE
        self.set_dest_filename()
        self.load_config()

    def load_config(self):
        self.source_config = ConfigFileLinuxIscsi(self.source_file)
        self.dest_config = ConfigFileLinuxIscsi(self.dest_filename)
        self.source_config.read()
        self.dest_config.read()

    def set_dest_filename(self):
        self.dest_filename = '/etc/iscsi.conf'


