#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import logging

from hostattachment.iscsicommon import test_iscsi_connectivity
from hostattachment.exceptions import HostAttachmentError
from hostattachment.platform.linux.execute import ExternalCommand
from hostattachment.platform.linux.iscsi.classes import IscsiQueryError
from hostattachment.platform.linux.iscsi import remove_bad_luns_from_all_iscsi_hosts
from hostattachment.platform.linux import messages

class ConnectionManager(object):
    def __init__(self):
#        self.log_used_classes()
        pass

#    def log_used_classes(self):
#        self.logger.log_a_class(IscsiAdm)

    def connect(self, ip_address, chap_name, chap_password):
        test_iscsi_connectivity(ip_address)
            
        iscsiadm = IscsiAdm(ip_address, chap_name, chap_password)
        if not iscsiadm.query():
            logging.error('Failed to query XIV array on IP Address %s' % ip_address)
            return False
        iscsiadm.discover_from_all_portals()
        iscsiadm.connect_to_all_portals()
        return True
        
    def disconnect(self, ip_address):
        iscsiadm = IscsiAdm(ip_address, None, None)
        try:
            test_iscsi_connectivity(ip_address)
        except HostAttachmentError, e:
            # Special case: we have no connectivity but we want to remove
            logging.warning("Couldn't query portal, removing only " + \
                "%s, instead of all related IPs" % (ip_address))
            iscsiadm.disconnect_from_portal(ip_address)
            iscsiadm.undiscover_through_portal(ip_address)

        else:
            # Regular case, get all IPs by querying the portal
            iscsiadm.query()
            iscsiadm.disconnect_from_all_portals()
            iscsiadm.undiscover_from_all_portals()

    def is_configured(self):
        return True

    def print_all(self):
        self.print_discoveries()
        self.print_nodes()
        self.print_sessions()
        print ''

    def print_discoveries(self):
        print 'Discovered portals and targets:'
        discovery = ExternalCommand('iscsiadm -m discovery')
        discovery.execute()
        discovery.print_output()
        print ''

    def print_nodes(self):
        print 'Discovered portals and targets:'
        nodes = ExternalCommand('iscsiadm -m node')
        nodes.execute()
        nodes.print_output()
        print ''

    def print_sessions(self):
        print 'Current connected portals and targets:'
        sessions = ExternalCommand('iscsiadm -m session')
        sessions.execute()
        sessions.print_output()
        print ''

    @staticmethod
    def iscsi_scan():
        logging.info(messages.REMOVING_BAD_LUNS)
        remove_bad_luns_from_all_iscsi_hosts()
        logging.info(messages.ISCSI_DAEMON_RELOAD)
        command = ExternalCommand('iscsiadm -m node -R')
        command.execute()

class IscsiAdm(object):
    def __init__(self, query_ip, chap_name, chap_password):
        self.target_name = ''
        self.portals = []
        self.query_ip = query_ip
        self.chap_name = chap_name
        self.chap_password = chap_password

    def query(self):
        discovery = ExternalCommand('iscsiadm -m discovery -t sendtargets -p %s' % self.query_ip)
        discovery.execute()
        if not discovery.is_error_free() and not discovery.contains('already exists'):
            logging.error('Failed to query %s' % self.query_ip)
            raise IscsiQueryError('Failed to query %s' % self.query_ip)
        for line in discovery.output.splitlines():
            (target_ip, target_name) = line.strip().split(' ',1)
            self.target_name = target_name
            self.portals.append(target_ip.split(',',1)[0])
            if line.__contains__('Target'):
                self.target_name = line.strip().split(' ')[1]
            elif line.__contains__('Portal'):
                self.portals.append(line.strip().split(':')[1].strip())
        if not self.portals.__contains__(self.query_ip):
            self.portals.append(self.query_ip)
        return True

    def discover_from_all_portals(self):
        logging.info('discovering iSCSI Portals')
        for portal in self.portals:
            self.discover_through_portal(portal)

    def undiscover_from_all_portals(self):
        logging.info('undiscovering iSCSI Portals')
        for portal in self.portals:
            self.undiscover_through_portal(portal)

    def connect_to_all_portals(self):
        logging.info('connecting to all iSCSI Portals')
        for portal in self.portals:
            self.connect_to_portal(portal)

    def disconnect_from_all_portals(self):
        logging.info('disconnecting from all iSCSI Portals')
        for portal in self.portals:
            self.disconnect_from_portal(portal)

    def connect_to_portal(self, portal):
        self.set_portal_auto_login(portal)
        self.set_chap_parameters(portal)
        self.login_to_portal(portal)
        
    def set_portal_auto_login(self, portal):
        connect = ExternalCommand('iscsiadm -m node -p %s -T %s -o update -n node.conn[0].startup -v automatic' % (portal, self.target_name))
        connect.execute()
        if not connect.is_error_free() and not connect.contains('already exists'):
            return False
        command = ExternalCommand('iscsiadm -m node -p %s -T %s -o update -n node.startup -v automatic' % (portal, self.target_name))
        command.execute()
        if not command.is_error_free() and not command.contains('already exists'):
            return False
        
    def set_chap_parameters(self, portal):
        if self.chap_name is None or self.chap_password is None:
            return None
        self.set_chap_name(portal)
        self.set_chap_password(portal)
        
    def set_chap_name(self, portal):
        connect = ExternalCommand('iscsiadm -m node -p %s -T %s -o update -n node.session.auth.username -v %s' % \
                                  (portal, self.target_name, self.chap_name))
        connect.execute()
        logging.debug('connect returned %s' % connect)
        if not connect.is_error_free():
            return False
        return True
    
    def set_chap_password(self, portal):
        connect = ExternalCommand('iscsiadm -m node -p %s -T %s -o update -n node.session.auth.password -v %s' % \
                                  (portal, self.target_name, self.chap_password))
        connect.execute()
        logging.debug('connect returned %s' % connect)
        if not connect.is_error_free():
            return False
        return True

    def login_to_portal(self, portal):
        command = ExternalCommand('iscsiadm -m node -p %s -T %s -l' % (portal, self.target_name))
        command.execute()
        if not command.is_error_free() and not command.contains('already exists'):
            logging.error('Failed to login to portal %s' % portal)
            raise HostAttachmentError("Failed to login to portal %s" % (portal))
        return True

    def disconnect_from_portal(self, portal):
        self.logout_from_portal(portal)
        self.remove_connection_to_portal(portal)

    def logout_from_portal(self, portal):
        logout = ExternalCommand('iscsiadm -m node -p %s -u' % (portal))
        logout.execute()

    def remove_connection_to_portal(self, portal):
        delete = ExternalCommand('iscsiadm -m node -p %s -o delete' % (portal))
        delete.execute()

    def discover_through_portal(self, portal):
        discovery = ExternalCommand('iscsiadm -m discovery -t sendtargets -p %s' % portal)
        discovery.execute()
        if not discovery.is_error_free() and not discovery.contains('already exists'):
            logging.warn('failed to query %s' % portal)

    def undiscover_through_portal(self, portal):
        discovery = ExternalCommand('iscsiadm -m discovery -p %s -o delete' % portal)
        discovery.execute()
        if not discovery.is_error_free() and not discovery.contains('already exists'):
            logging.warn('failed to remove discovery address %s' % portal)
            return False
        return True

