#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

#
# multipath.conf parser, verifier and configurator
# Terminology:
# REQUIRED means: a configuration directive that needs to be in multipath.conf
# CURRENT means : stuff from the parsed multipath.conf (current configuration)
#
# e.g., devices_required vs. devices_current
#

import re
import time
import os
from shutil import copyfile
import logging

from hostattachment import definitions
from hostattachment.exceptions import HostAttachmentError
from hostattachment.classes import AttachmentConfiguration
from hostattachment.platform.linux import messages
from hostattachment.platform.linux.common import generate_backup_filename

MULTIPATH_CONF_PATH = "/etc/multipath.conf"

def nothing_starts_with(mystring, prefixes):
    """A helper function that returns True if mystring doesn't start with any
    prefix in a list of prefixes"""
    for prefix in prefixes:
        if mystring.startswith(prefix):
            return False

    return True

# Required configuration parameters for multipath.conf
required_config = {}
required_config["defaults"] = { "user_friendly_names": "yes" }
required_config["blacklist_devnodes"] = [ '"^(ram|raw|loop|fd|md|dm-|sr|scd|st)[0-9]*"',
                       '"^hd[a-z]"', '"^cciss!c[0-9]d[0-9]*"' ]
required_config["blacklist_vendors"] = [ '"IBM-ESXS"', '"LSILOGIC"', '"ATA"', '"VMware"' ]
required_config["devices"] = [ \
            { "vendor": ['"IBM"'], "product": ['"2810XIV"'],
              "path_selector": ['"round-robin 0"'],
              "path_grouping_policy": ["multibus"], "path_checker": ["tur"],
              "polling_interval": ["3"], "failback": ["15"], "rr_min_io": ["15"],
              "no_path_retry": ["5"] },
            { "vendor": ['"XIV"'], "product": ['"NEXTRA"'],
              "path_selector": ['"round-robin 0"'],
              "path_grouping_policy": ["multibus"], "path_checker": ["tur"],
              "polling_interval": ["3"], "failback": ["15"], "rr_min_io": ["15"],
              "no_path_retry": ["5"] },
        ]
required_config["devnodes_banned_from_blacklist"] = ['"*', '"^*', '"sd', '"^sd' ]
required_config["wwids_banned_from_blacklist"] = ['"*']

class MultipathConf(object):
    """Handles verifying and configuring the multipath.conf file"""
    def __init__(self, filename = "/etc/multipath.conf"):
        super(MultipathConf, self).__init__()
        try:
            contents = file(filename).read()
        except IOError, e:
            # Could be many things, we act as if the file is empty
            logging.warning(messages.FAILED_READING_FILE % (filename))
            contents = ""

        self.filename = filename
        attachment_configuration = AttachmentConfiguration.getConfiguration()
        self.backup_path = attachment_configuration.backup_path
        self.config = generic_parse_config(contents)

    def is_configured(self):
        result = self._verify_not_empty() and \
                 self._verify_defaults() and \
                 self._verify_blacklist() and \
                 self._verify_devices()

        return result

    def configure(self):
        self._config_skeleton()
        self._config_defaults()
        self._config_blacklist()
        self._config_devices()
        new_file_contents = self._header()
        new_file_contents += unparse_config(self.config)
        self._backup()
        self._dump_to_file(new_file_contents)

    def _header(self):
        (tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec, tm_wday, tm_yday, tm_isdst) = time.localtime()
        header = '# Linux Multipath I/O Configuration File\n' + \
                 '# Version %s\n' % (definitions.HOSTATTACHMENT_VERSION) + \
                 '# Generated by IBM-XIV Host Attachment at %s-%s-%s %s:%s:%s\n' % (tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec) + \
                 '# Do not manually modify this file!\n\n'

        return header

    def _backup(self):
        if os.path.exists(self.filename):
            backup_file = generate_backup_filename(self.backup_path,
                                    os.path.basename(self.filename))
            logging.info(messages.BACKING_UP_SRC_TO_DST % (self.filename, backup_file))
            copyfile(self.filename, backup_file)

    def _dump_to_file(self, contents):
        f = file(self.filename, "w")
        f.write(contents)
        f.close()

    def _verify_not_empty(self):
        # Empty
        if self.config == {}:
            logging.error(messages.MULTIPATH_CONF_IS_EMPTY)
            return False

        return True

    def _verify_defaults(self):
        try:
            for key in required_config["defaults"].keys():
                logging.debug("multipath.conf: verifying defaults: %s" % (key))
                value = required_config["defaults"][key]
                assert(self.config["defaults"][0][key] == [value])
        except (AssertionError, KeyError), e:
            logging.error("multipath.conf: defaults does not contain (%s:%s)" % (key, value))
            return False

        return True

    def _verify_blacklist(self):
        logging.debug("multipath.conf: verifying blacklist")
        try:
            blacklist_devnodes_required = set(required_config["blacklist_devnodes"])
            blacklist_devnodes_current = set(self.config["blacklist"][0]["devnode"])

            blacklist_vendors_required = \
                    set(required_config["blacklist_vendors"])
            blacklist_vendors_current = set([ device["vendor"][0] \
                    for device in self.config["blacklist"][0]["device"] ])
        except (KeyError), e:
            logging.error("multipath.conf: missing fields in the blacklist block")
            return False

        try:
            blacklist_wwids_current = self.config["blacklist"][0]["wwid"]
        except KeyError, e:
            blacklist_wwids_current = []

        try:
            legacy_devnode_blacklist_current = self.config["devnode_blacklist"][0]["devnode"]
        except KeyError, e:
            legacy_devnode_blacklist_current = []

        # Makes sure that required blacklisted devnodes are included 
        # in the currently configured devnodes
        if not blacklist_devnodes_required.issubset(blacklist_devnodes_current):
            logging.error("multipath.conf: required blacklist values are missing")
            return False

        # Makes sure that blacklisted vendors are included in currently
        # configured blacklisted vendors (device { } section)
        if not blacklist_vendors_required.issubset(blacklist_vendors_current):
            logging.error("multipath.conf: required blacklist values are missing")
            return False

        # Make sure blacklist does not contain our stuff (devices + devnodes)
        try:
            devices_vendors_current = set([device["vendor"][0] for device in self.config["devices"][0]["device"]])
        except (KeyError), e:
            return False
        if not devices_vendors_current.isdisjoint(blacklist_vendors_current):
            logging.error("multipath.conf: blacklist contains items that mustn't be there")
            return False
        for blacklist_devnode_current in blacklist_devnodes_current:
            for banned_devnode in required_config["devnodes_banned_from_blacklist"]:
                if blacklist_devnode_current.startswith(banned_devnode):
                    logging.error("multipath.conf: blacklist contains items that mustn't be there")
                    return False
        for blacklist_wwid_current in blacklist_wwids_current:
            for banned_wwid in required_config["wwids_banned_from_blacklist"]:
                if blacklist_wwid_current.startswith(banned_wwid):
                    logging.error("multipath.conf: blacklist contains items that mustn't be there")
                    return False
        for devnode_blacklist_current in legacy_devnode_blacklist_current:
            for banned_devnode in required_config["devnodes_banned_from_blacklist"]:
                if devnode_blacklist_current.startswith(banned_devnode):
                    logging.error("multipath.conf: blacklist contains items that mustn't be there")
                    return False

        return True

    def _verify_devices(self):
        # DEVICES
        for device_required in required_config["devices"]:
            logging.debug("multipath.conf: asserting device %s" % device_required["product"][0])
            try:
                devices_current = self.config["devices"][0]["device"]
                assert(device_required in devices_current)
            except (KeyError, AssertionError), e:
                logging.error("multipath.conf: required devices are missing")
                return False

        return True

    def _config_skeleton(self):
        """Create skeleton if doesn't exist"""
        if not self.config.has_key("defaults"):
            self.config["defaults"] = [{}]
        if not self.config.has_key("blacklist"):
            self.config["blacklist"] = [{}]
        if not self.config.has_key("devices"):
            self.config["devices"] = [{}]

    def _config_defaults(self):
        for key in required_config["defaults"].keys():
            value = required_config["defaults"][key]
            if key not in self.config["defaults"][0] or self.config["defaults"][0][key] != [value]:
                logging.info("multipath.conf: Adding default: %s:%s" % (key, value))
                self.config["defaults"][0][key] = [value]

    def _config_blacklist(self):
        if not self.config["blacklist"][0].has_key("devnode"):
            self.config["blacklist"][0]["devnode"] = []

        if not self.config["blacklist"][0].has_key("device"):
            self.config["blacklist"][0]["device"] = []

        for value in required_config["blacklist_devnodes"]:
            if value not in self.config["blacklist"][0]["devnode"]:
                logging.info("multipath.conf: adding blacklist: devnode %s" % value)
                self.config["blacklist"][0]["devnode"].append(value)

        blacklisted_vendors_current = []
        for device in self.config["blacklist"][0]["device"]:
            try:
                blacklisted_vendors_current.append(device["vendor"][0])
            except KeyError, e:
                pass

        for blacklist_vendor_required in required_config["blacklist_vendors"]:
            if blacklist_vendor_required not in blacklisted_vendors_current:
                logging.info("multipath.conf: adding blacklist: vendor %s" % blacklist_vendor_required)
                self.config["blacklist"][0]["device"].append({"vendor": [blacklist_vendor_required]})

        # Remove ourselves from blacklist (devices + devnodes)
        try:
            vendors_current = [device["vendor"][0] for device in self.config["devices"][0]["device"]]
        except KeyError: # no vendors found
            pass
        else:
            logging.info("multipath.conf: filtering bad values from blacklist (device)...")
            self.config["blacklist"][0]["device"] = \
                filter(lambda x: x["vendor"][0] not in vendors_current, self.config["blacklist"][0]["device"])

        try:
            blacklist_devnodes_current = self.config["blacklist"][0]["devnode"]
        except KeyError: # no devnodes found
            pass
        else:
            logging.info("multipath.conf: filtering bad values from blacklist (devnode)...")
            self.config["blacklist"][0]["devnode"] = \
                filter(lambda x: nothing_starts_with(x, required_config["devnodes_banned_from_blacklist"]),
                       blacklist_devnodes_current)

        try:
            blacklist_wwids_current = self.config["blacklist"][0]["wwid"]
        except (KeyError), e:
            pass
        else:
            logging.info("multipath.conf: filtering bad values from blacklist (wwid)...")
            self.config["blacklist"][0]["wwid"] = \
                filter(lambda x: nothing_starts_with(x, required_config["wwids_banned_from_blacklist"]),
                       blacklist_wwids_current)

        try:
            legacy_devnode_blacklist_current = self.config["devnode_blacklist"][0]["devnode"]
        except (KeyError), e:
            pass
        else:
            logging.info("multipath.conf: filtering bad values from blacklist (legacy devnode)...")
            self.config["devnode_blacklist"][0]["devnode"] = \
                filter(lambda x: nothing_starts_with(x, required_config["devnodes_banned_from_blacklist"]),
                       legacy_devnode_blacklist_current)



    def _config_devices(self):
        """Configures the devices block. Policy: if product is already there
        (e.g. 2810XIV), don't mess with its settings. If it's not, create a new
        one."""
        if not self.config["devices"][0].has_key("device"):
            self.config["devices"][0]["device"] = []

        devices_current = self.config["devices"][0]["device"]

        # Make a list of product+vendor (which we use as a key) that are set in the current multipath.conf
        device_items_current = []
        for device in self.config["devices"][0]["device"]:
            try:
                 device_items_current.append( (device["product"][0], device["vendor"][0]) )
            except KeyError, e:
                 pass

        for device_required in required_config["devices"]:
            product_required = device_required["product"][0]
            vendor_required = device_required["vendor"][0]
            # CASE #1: XIV device isn't there at all
            # Action: add a new entry
            if (product_required, vendor_required) not in device_items_current:
                logging.info("multipath.conf: adding device %s:%s" % (product_required, vendor_required))
                self.config["devices"][0]["device"].append(device_required)
            # CASE #2: XIV device is defined, but with the wrong parameters
            # Action: replace the entry with the predefined "required" entry
            elif device_required not in devices_current: # no perfect match. something differs.
                logging.info("multipath.conf: updating device: product %s" % product_required)
                for i in range(len(self.config["devices"][0]["device"])):
                    if self.config["devices"][0]["device"][i]["product"][0] == product_required and \
                       self.config["devices"][0]["device"][i]["vendor"][0] == vendor_required:
                        self.config["devices"][0]["device"][i] = device_required


    def _set_keyword_hierarchy(self):
        # This skeleton was built according to multipath-tools 0.4.8's source
        # Might be useful someday to validate input/output
        self.keywords = {
              "defaults": [ "polling_interval", "udev_dir", "selector",
                            "path_grouping_policy", "getuid_callout",
                            "prio_callout", "features", "path_checker",
                            "failback", "rr_min_io", "rr_weight",
                            "no_path_retry", "pg_timeout",
                            "user_friendly_names",
                          ],
              "blacklist": { "devnode": None, "wwid": None,
                             "device": [ "vendor", "product" ],
                           },
              "blacklist_exceptions": { "devnode": None, "wwid": None,
                             "device": [ "vendor", "product" ],
                           },
              "devices": { "device": [ "vendor", "product", "product_blacklist",
                                        "path_grouping_policy", "getuid_callout",
                                        "path_selector", "path_checker", "features",
                                        "hardware_handler", "prio_callout",
                                        "failback", "rr_weight", "no_path_retry",
                                        "rr_min_io", "pg_timeout" ],
                        },
              "multipaths": { "multipath": [ "wwid", "alias",
                                             "path_grouping_policy",
                                             "path_selector",
                                             "failback", "rr_weight",
                                             "no_path_retry", "rr_min_io",
                                             "pg_timeout",
                                           ],
                            },
            }




def generic_parse_config(data):
    """Parses a simple config file (multipath.conf format), that contains two data types:
    - Type1: key <space> value
    - Type2: block { <data> } (<data> may contain either Type1's and/or Type2's)

    Returns a datatype representing this configuration
    """
    pos = 0
    depth = 0
    parsed_data = {}
    context = parsed_data
    stack = []
    while (pos < len(data)):
        # Type1 regexps
        keyvalue = re.compile("^\s*(\w+)\s+(.+)\s*\n").search(data[pos:])
        keyvalue_quoted = re.compile("^\s*(\w+)\s+(\".+\")\s*\n").search(data[pos:])

        # Type2 regexps
        block = re.compile("^\s*(\w+)\s*{\s*\n").search(data[pos:])
        block_end = re.compile("^\s*}\s*\n").search(data[pos:])
        comment = re.compile("^#.*\n").search(data[pos:])
        empty = re.compile("^\s*\n").search(data[pos:])
        line = re.compile("^(.*)\n").search(data[pos:])
        if empty is not None:
            pos += empty.end()
        elif block is not None:
            # Type2 was detected
            depth += 1
            block_name = block.groups()[0]
            logging.debug("Block %s, depth %d" % (block_name, depth))
            try:
                context[block_name].append({})
            except KeyError, e:
                context[block_name] = []
                context[block_name].append({})
            last = len(context[block_name])
            stack.append(context)
            context = context[block_name][last - 1]
            pos += block.end()
        elif keyvalue is not None or keyvalue_quoted is not None:
            # Type1 was detected
            if keyvalue_quoted is not None:
                kv = keyvalue_quoted
            else:
                kv = keyvalue
            key = kv.groups()[0]
            value = kv.groups()[1]
            logging.debug("KeyValue %s:%s" % (key, value))
            try:
                context[key].append(value)
            except KeyError, e:
                context[key] = []
                context[key].append(value)
            pos += kv.end()
        elif block_end is not None:
            depth -= 1
            context = stack.pop()
            logging.debug("Block end, depth %d" % (depth))
            pos += block_end.end()
        elif comment is not None:
            pos += comment.end()
        else:
            raise Exception('Bad line: "%s"' % (line.groups()[0]))

    return parsed_data

def unparse_config(config, level = 0):
    """Internal configuration datatype to multipath.conf format"""
    output = ""
    for key in config.keys():
        for key_instance in range(len(config[key])):
            if isinstance(config[key][key_instance], dict):
                output += "%s%s {\n" % (level*" ", key)
                output += unparse_config(config[key][key_instance], level + 4)
                output += "%s}\n" % (level*" ")
            else:
                output += "%s%s %s\n" % (level*" ", key, config[key][key_instance])

    return output
