#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import time
import os
import logging
from shutil import copyfile

from hostattachment.platform.linux import definitions as linux_hak_defs
from hostattachment.platform.linux import messages
from hostattachment.platform.linux.common import generate_backup_filename
from hostattachment.exceptions import HostAttachmentError
from hostattachment.platform.linux.classes import get_os_string
from hostattachment.classes import AttachmentConfiguration

from xiv_xsf.host.os.linux.udev import udev_trigger, UdevError

UDEV_RULES_PATH = os.path.join(os.path.sep, "etc", "udev", "rules.d")

class UdevRules(object):
    def __init__(self):
        self.os_string = get_os_string()
        self.rules_source = linux_hak_defs.UDEV_RULES_SOURCE
        self.udev_rules = []
        self.udev_rules.append('40-multipath.rules')
        self.udev_rules.append('44-ibm-xiv.rules')
        attachment_configuration = AttachmentConfiguration.getConfiguration()
        self.backup_path = attachment_configuration.backup_path

    def is_configured(self):
        return self.dst_rules_configured()

    def _get_src_file(self, dir, filename):
        """Tries all optional filenames in an array and returns the first
        match. In this case, try os-specific first and then generic version"""
        options = []
        options.append('%s/%s.%s' % (dir, filename, self.os_string))
        options.append('%s/%s' % (dir, filename))

        for option in options:
            if (os.path.exists(option)):
                return option

        raise HostAttachmentError(messages.CANT_FIND_ANY_FILE % (options) )

    def rules_differ(self, rules1, rules2):
        """Compares the two rules file contents"""
        # TODO should be a smarter compare (i.e. ignore comments and spaces)
        if rules1 != rules2:
            return True

        return False

    def dst_rules_configured(self):
        for filename in self.udev_rules:
            src_filename = self._get_src_file(self.rules_source, filename)
            src = file(src_filename)
            try:
                dst = file('%s/%s' % (UDEV_RULES_PATH, filename))
            except IOError, e:
                logging.error(messages.UDEV_RULES_FILE_DOESNT_EXIST % (filename))
                return False

            if self.rules_differ(src.read(), dst.read()):
                logging.error(messages.UDEV_RULE_NOT_CONFIGURED % (filename))
                return False

        return True

    def backup(self, src):
        if (os.path.exists(src)):
            backup_file = generate_backup_filename(self.backup_path,
                                            os.path.basename(src))
            logging.info(messages.BACKING_UP_SRC_TO_DST % (src, backup_file))
            copyfile(src, backup_file)

    def configure(self):
        for basename in self.udev_rules:
            src_filename = self._get_src_file(self.rules_source, basename)
            dst_filename = os.path.join(UDEV_RULES_PATH, basename)
            self.backup(dst_filename)
            try:
                copyfile(src_filename, dst_filename)
                logging.info(messages.UDEV_RULE_CONFIG_SUCCESS % (basename))
            except IOError:
                raise HostAttachmentError(messages.UDEV_RULE_CONFIG_FAILURE % (basename) )
        try:
            udev_trigger()
        except UdevError:
            # This is usually okay
            logging.warning(messages.UDEV_TRIGGER_FAILED)
        else:
            logging.info(messages.UDEV_TRIGGERED)
