#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import re
import logging

from hostattachment.platform.linux.execute import ExternalCommand
from hostattachment.platform.linux import definitions as linux_hak_defs

class HctlNotFound(Exception):
    pass

def extract_hctl_from_line(line):
    line_splits = line.split(' ')
    host = line_splits[1].strip('scsi')
    channel = remove_leading_zeros(line_splits[3].strip())
    target = remove_leading_zeros(line_splits[5].strip())
    lun = remove_leading_zeros(line_splits[7].strip())
    return (host, channel, target, lun)

def remove_leading_zeros(string):
    non_zeros = re.sub('^0*', '', string)
    if non_zeros == '':
        return '0'
    return non_zeros

def is_lun_zero(hctl):
    if hctl.split().pop() == '0':
        return True
    else:
        return False

def remove_luns(devices_list):
    # TODO can we use /sys/block/device/dev/delete ? supported on all platforms?
    for device in devices_list:
        logging.info('Removing single device %s' % device)
        remove_single_device_command = ExternalCommand(\
            '''echo 'scsi remove-single-device %s' > /proc/scsi/scsi''' % (device))
        remove_single_device_command.execute()

def hctl_should_be_removed(hctl):
    try:
        device_node = get_device_by_hctl(hctl)
    except HctlNotFound:
        return False # We cannot handle this

    if is_lun_zero(hctl):
        return False # For XIV, we do not remove lun0 devices
                     # TODO what about other vendors?
    if not is_device_responding(device_node):
        return True

    return False

def is_device_responding(device):
    sg_turs = ExternalCommand('sg_turs %s' % (device))
    sg_turs.execute()
    if sg_turs.is_error_free():
        return True
    if sg_turs.contains('with 0 errors'):
        return True
    return False

def get_device_by_hctl(hctl):
    """Gets a string-hctl (e.g. '1 2 3 4') and looks for a matching /dev/sg*
    device.
    Returns the device full path if found, None otherwise"""
    sg_map = ExternalCommand('sg_map -x')
    sg_map.execute()
    for line in sg_map.output.splitlines():
        if "Strange" in line:
            logging.warning('sg_map reported strange errors')
            continue # Try next line

        regexp_string = "^(/dev/sg\d+)\s+"
        for member in hctl.split(" "):
            regexp_string += "%s\s+" % (member)
        match = re.compile(regexp_string).search(line)
        if match is not None:
            return match.groups()[0]

    raise HctlNotFound(hctl)

def get_hctls_by_host(host_number):
    hctl_list = []
    proc_scsi_scsi = file("/proc/scsi/scsi").readlines()
    for line in proc_scsi_scsi:
        if line.__contains__('Host: scsi%s' % host_number):
            (host, channel, target, lun) = extract_hctl_from_line(line)
            logging.debug('Found an XIV LUN: %s %s %s %s' % (host, channel, target, lun))
            hctl_list.append('%s %s %s %s' % (host, channel, target, lun))
    return hctl_list

def get_hctls_to_remove(host_number):
    all_hctls_of_host = get_hctls_by_host(host_number)
    hctls_to_remove = filter(hctl_should_be_removed, all_hctls_of_host)

    return hctls_to_remove


def get_targets_from_lun_zero(host_number):
    hct_list = []
    proc_scsi_scsi = file('/proc/scsi/scsi').readlines()
    for line in proc_scsi_scsi:
        if line.__contains__('Host: scsi%s' % host_number):
            (host, channel, target, lun) = extract_hctl_from_line(line)
            if lun == '0' and not hct_list.__contains__('%s %s %s' % (host, channel, target)):
                logging.debug('found an XIV target: %s %s %s %s' % (host, channel, target, lun))
                hct_list.append('%s %s %s' % (host, channel, target))
    return hct_list

def send_scan_to_scsi_host(host_number):
    driver_scan_command = ExternalCommand('''echo '- - -' > /sys/class/scsi_host/%s/scan''' % (host_number))
    driver_scan_command.execute()
    if driver_scan_command.return_value is 0:
        return True
    return False

def get_scsi_host_number(scsi_host):
    host_number = scsi_host.replace('host', '')
    return host_number

def remove_bad_luns(scsi_host):
    host_number = get_scsi_host_number(scsi_host)
    hctls_to_remove = get_hctls_to_remove(host_number)
    remove_luns(hctls_to_remove)

def scan_by_add_single_device(host_number):
    for hct in get_targets_from_lun_zero(host_number):
        logging.debug('scanning target %s for new LUNs' % hct)
        for lun in range(0, linux_hak_defs.MAX_LUN_SCAN):
            add_single_device = ExternalCommand(\
                '''echo 'scsi add-single-device %s %s' > /proc/scsi/scsi''' % (hct, lun))
            add_single_device.execute()

def add_new_luns(scsi_host):
    if not send_scan_to_scsi_host(scsi_host):
        host_number = get_scsi_host_number(scsi_host)
        scan_by_add_single_device(host_number)
