#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import logging

from hostattachment.platform.linux.execute import ExternalCommand
from hostattachment.exceptions import HostAttachmentError
from hostattachment.platform.linux import messages

class GenericServiceManager(object):
    def __init__(self, init_script, daemon_name):
        self.init_script = init_script
        self.daemon_name = daemon_name

    def status(self):
        logging.debug('Checking status of the %s service' % self.init_script)
        proc = ExternalCommand('/etc/init.d/%s status' % self.init_script)
        proc.execute()
        return proc.return_value

    def start(self, ignore_retval = False):
        logging.debug('Starting the %s service' % self.init_script)
        proc = ExternalCommand('/etc/init.d/%s start' % self.init_script)
        proc.execute()
        if not ignore_retval and proc.return_value != 0:
            raise HostAttachmentError("Failed starting the " + \
                "\"%s\" service" % (self.init_script))

    def restart(self, ignore_retval = False):
        logging.debug('Restarting the %s service' % self.init_script)
        proc = ExternalCommand('/etc/init.d/%s restart' % self.init_script)
        proc.execute()
        if not ignore_retval and proc.return_value != 0:
            raise HostAttachmentError("Failed restarting the " + \
                "\"%s\" service" % (self.init_script))

    def reload(self, ignore_retval = False):
        logging.debug('Reloading the %s service' % self.init_script)
        proc = ExternalCommand('/etc/init.d/%s reload' % self.init_script)
        proc.execute()
        if not ignore_retval and proc.return_value != 0:
            raise HostAttachmentError("Failed reloading the " + \
                "\"%s\" service" % (self.init_script))

    def is_configured(self):
        if not self._is_chkconfig_set():
            logging.warning(messages.SERVICE_NOT_ENABLED % (self.init_script))
            return False

        if not self._is_daemon_running() or self.status() != 0:
            logging.warning(messages.DAEMON_NOT_RUNNING % (self.daemon_name))
            return False

        return True

    def _is_chkconfig_set(self):
        ps_command = ExternalCommand('chkconfig --list %s' % self.init_script)
        ps_command.execute()
        if ps_command.return_value == 0:
            for runlevel in [3, 4, 5]:
                if not ps_command.output.__contains__("%d:on" % runlevel):
                    return False # Service is off for at least one runlevel
        else:
            return False # Service is not registered in chkconfig

        return True

    def _is_daemon_running(self):
        ps_command = ExternalCommand('/sbin/pidof %s' % (self.daemon_name))
        ps_command.execute()

        return ps_command.return_value == 0

    def configure(self, ignore_service_retval = False):
        """Configures the service and restarts it.
        Will not check the retval of init start if ignore_service_retval = True
        """
        add_command = ExternalCommand('chkconfig --add %s' % self.init_script)
        add_command.execute()
        if add_command.return_value != 0:
            raise HostAttachmentError(messages.FAILED_REGISTERING_SERVICE % (self.init_script))

        # chkconfig is a RHEL script, but SLES supply a compatibility script
        # which seems to work very well
        on_command = ExternalCommand('chkconfig --level 345 %s on' % self.init_script)
        on_command.execute()
        if add_command.return_value != 0:
            raise HostAttachmentError(messages.FAILED_ENABLING_SERVICE % (self.init_script, "3,4,5"))

        # Start if needed
        if self.status() != 0:
            self.start(ignore_service_retval)
