#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 

from xiv_xsf.host.storage.multipath import MultipathDevice, MultipathInfoError,\
    get_all_multipath_devices, get_all_non_multipath_devices,\
    MP_NATIVE, MP_VERITAS, MP_AUTO
from xiv_xsf.host.storage import get_all_storage_devices
from xiv_xsf.scsy.exceptions import ScsiStatusCheckCondition, CannotOpenDeviceError
from xiv_xsf.scsy import raw as scsy_raw
from xiv_xsf.scsy import commands as scsy_commands
from xiv_xsf.console import print_in_color, get_console_dimensions
from excpts import exception_catcher
from construct.core import FieldError

def format_size(size_in_bytes):
    units = [ 'B', 'KB', 'MB', 'GB', 'TB' ]
    unit_no = 0

    new_size = float(size_in_bytes)
    for i in range(len(units)):
        if (new_size / 1000 > 1):
            new_size = new_size / 1000
        else:
            break

    return "%.1f%s" % (round(new_size, 1), units[i])

def dict_to_sorted_list(my_dict):
    my_sorted_list = []
    keys = my_dict.keys()
    keys.sort()
    for key in keys:
        my_sorted_list.append(my_dict[key])

    return my_sorted_list

class DevListFieldHandler(object):
    """Holds static functions that return device specific field"""

    def __init__(self):
        self.scsy_device = scsy_raw.RawDevice()

    def prefetch(self):
        """Pre fetches some info we'll need later, instead of reading it for
        each field separately"""
        pass

    def get_device(self, disk_device):
        return disk_device.name

    def get_xiv_id(self, disk_device):
        return disk_device.xiv.machine_serial_number_decimal

    def get_vol_id(self, disk_device):
        return disk_device.xiv.volume_serial_number_decimal
    
    def get_xiv_id_hex(self, disk_device):
        return disk_device.xiv.machine_serial_number

    def get_vol_id_hex(self, disk_device):
        return disk_device.xiv.volume_serial_number

    def get_vol_name(self, disk_device):
        """Returns the volume name"""
        return disk_device.xiv.volume_name

    def get_vendor(self, disk_device):
        """Returns the vendor string"""
        return disk_device.scsi_standard_inquiry.vendor_identification

    def get_xiv_host(self, disk_device):
        """Returns the hostname as configured in xiv storage system"""
        return disk_device.xiv.host
    
    def get_is_xiv_device(self, disk_device):
        return disk_device.is_xiv_device

    def get_size(self, disk_device):
        """Returns the capacity in bytes of given device"""
        capacity = None

        self.scsy_device.openHandle(disk_device.device.raw_device_filename)
        cmd = scsy_commands.ReadCapacity16()
        try: # Try 16
            capacity_data = cmd.basic(self.scsy_device)
        except IOError, e: # CHK : SCSY error got changed?
            cmd = scsy_commands.ReadCapacity10()

            try: # Try 10
                capacity_data = cmd.basic(self.scsy_device)
            except Exception, e:
                capacity = 0

        except FieldError, e:
            capacity = 0
        except ScsiStatusCheckCondition, e:
            capacity = 0
        finally:
            self.scsy_device.closeHandle()

        if capacity != 0:
            capacity = capacity_data.returned_logical_block_address * \
                capacity_data.logical_block_length_in_bytes
            formatted_capacity = format_size(capacity)
        else:
            formatted_capacity = "N/A"

        return formatted_capacity

    def get_multipath(self, disk_device):
        # Applicable only for multipath devices
        if not isinstance(disk_device, MultipathDevice):
            return "N/A"

        try:
            return_string = "%d" % (disk_device.active_paths)
            return_string += "/%d" % (disk_device.total_paths)
        except KeyError, e:
            return_string = "N/A"
        except MultipathInfoError, e:
            return_string = "N/A"

        return return_string

    def get_lun(self, disk_device):
        return_string = str(disk_device.lun)
        return return_string

    def get_serial(self, disk_device):
        return str(disk_device.scsi_serial)

class DevListData(object):
    
    def __init__(self, fields, mp_framework = MP_AUTO):
        self.fields = fields
        self.field_handler = DevListFieldHandler()
        self.all_devices = { "xiv": {}, "nonxiv": {} }
        self.errored_devices = []
        self.scsy_device = scsy_raw.RawDevice()
        self.mp_framework = mp_framework

    @exception_catcher
    def populate_data(self):
        all_disk_devices = get_all_storage_devices(mp_framework_type = self.mp_framework)

        for disk_device in all_disk_devices:
            try:
                self.scsy_device.openHandle(disk_device.device.raw_device_filename)
                cmd = scsy_commands.TestUnitReady()
                try:
                    cmd.basic(self.scsy_device)
                finally:
                    self.scsy_device.closeHandle()

                if disk_device.is_xiv_device:
                    self.all_devices["xiv"][disk_device.device.filename] = \
                        self.get_device_data(disk_device)
                else: # non-xiv
                    self.all_devices["nonxiv"][disk_device.device.filename] = \
                        self.get_device_data(disk_device)

            except CannotOpenDeviceError, e:
                self.errored_devices.append(disk_device.device.filename)
                continue
            
            except Exception, e:
                # We'd like to RAISE this exception in debug mode
                self.errored_devices.append(disk_device.device.filename)

    def get_device_data(self, device):
        line = ""
        device_data = {}

        self.field_handler.prefetch()

        for field in self.fields:
            value = None

            if (device.is_xiv_device) or (not field.xiv_specific):
                func = getattr(self.field_handler, "get_%s" % field.name)
                try:
                    device_data[field.name] = func(device)
                except Exception, e:
                    pass

        return device_data

    def mk_sorted_list(self):
        sorted_devices_list = {}

        sorted_devices_list["xiv"] = \
            dict_to_sorted_list(self.all_devices["xiv"])

        sorted_devices_list["nonxiv"] = \
            dict_to_sorted_list(self.all_devices["nonxiv"])

        self.sorted_devices_list = sorted_devices_list

