#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program  
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
# 
#-------------------------------------------------------------------------------------------------   

import os, time, zlib
from os import path
from zipfile import ZipFile, ZIP_DEFLATED
from xml.etree import ElementTree
from xml.etree.ElementTree import SubElement
from tempfile import mkdtemp, mktemp
from shutil import rmtree

from collectable.exceptions import CollectionError
from xiv_xsf.console import print_in_color

DEFAULT_DIR="/tmp"

class XIVCollector(object):
    def __init__(self, default_dir = DEFAULT_DIR):
        self.attributes = dict()
        self.xml_root = ElementTree.Element("XcollectItem")
        self.add_attribute('run-date', '%s-%s-%s_%s-%s-%s' % (time.localtime()[0:6]))
        self.default_dir = default_dir
        self.collectables = []
        self.tempdir = mkdtemp(prefix='xiv_diag_')

    def __str__(self):
        return 'XIVdiag, version %s' % self.version

    def mktemp(self, suffix='', prefix=''):
        return mktemp(suffix, prefix, self.tempdir)

    def add_attribute(self, name, value):
        if self.attributes.has_key(name):
            raise AttributeError("Attribute %s already exists" % name)
        self.attributes[name] = str(value)
        attribute = SubElement(self.xml_root, name)
        attribute.text = str(value)

    def get_attribute(self, name):
        if not self.attributes.has_key(name):
                raise AttributeError("Attribute %s does not exist" % name)
        return self.attributes[name]

    def get_zip_file(self):
        self.zip_file = None
        self.filename = 'xiv_diag-results_%s.zip' % (self.get_attribute('run-date'))
        while not self.zip_file:
            self.ask_for_directory()
            self.open_zip_file()

    def ask_for_directory(self):
        self.dirpath = raw_input('Please type in a path to place the xiv_diag file in [default: %s]: ' %  self.default_dir)
        if self.dirpath == '':
            self.dirpath = self.default_dir

    def open_zip_file(self):
        line = 'Creating xiv_diag zip file %s ' % path.sep.join([self.filename])
        print_in_color(line.ljust(70, ' '), newline=False)
        try:
            self.zip_file = ZipFile(file=path.sep.join([self.dirpath, self.filename]), mode='w', compression=ZIP_DEFLATED)
        except:
            print_in_color('ERROR'.rjust(9, ' '), 'red')
            printin_color('ERROR: failed to create %s in the %s, please choose another one.' % (self.filename, self.dirpath), 'dark_red')
        print ''

    def collect_collectable(self, collectable):
        line = 'INFO: Gathering %s...     ' % (collectable.get_attribute('description'))
        print_in_color(line.ljust(70,' '), newline=False)
        try:
            collectable.collect()
            print_in_color('DONE'.rjust(9, ' '), 'dark_green')
        except CollectionError, e:
            if str(e).__contains__('Errno 2') or str(e).__contains__('Error 2') or str(e).__contains__('Error 3'):
                print_in_color('SKIPPED'.rjust(9, ' '), 'yellow')
            else:
                print_in_color('FAILED'.rjust(9, ' '), 'red')
                print_in_color('ERROR: Could not gather %s:\n (%s)' \
                               % (collectable.get_attribute('description'), str(e)), 'dark_red')
        self.add_collectable_elements_to_xml(collectable)

    def add_collectable_elements_to_xml(self, collectable):
        collectable_element = SubElement(self.xml_root, 'CollectableItem')
        for name, value in collectable.get_attributes_dictionary().iteritems():
            attribute = SubElement(collectable_element, name)
            attribute.text = str(value)

    def collect(self):
        for collectable in self.collectables:
            self.collect_collectable(collectable)
        self.add_xml_to_zipfile()

    def add_xml_to_zipfile(self):
        self.zip_file.writestr('xiv_diag_info.xml', ElementTree.tostring(self.xml_root))

    def close(self):
        line = 'INFO: Closing xiv_diag zip file'
        print_in_color('\n' + line.ljust(70, ' '), newline=False)
        self.zip_file.close()
        print_in_color('DONE'.rjust(9, ' '), 'dark_green')
        line =  'Deleting temporary directory...'.ljust(70,' ')
        print_in_color(line.ljust(70, ' '), newline=False)
        rmtree(self.tempdir, ignore_errors=True)
        print_in_color('DONE'.rjust(9, ' '), 'dark_green')
        print 'INFO: Gathering is now complete.'
        print 'INFO: You can now send', path.sep.join([self.dirpath, self.filename]), 'to IBM-XIV for review.'
        print 'INFO: Exiting.'
    
    def test(self):
        test_zip_file = ZipFile(file=path.sep.join([self.dirpath, self.filename]), mode='r')
        return test_zip_file.testzip()

    def add_collectable_item(self, collectable_item):
        self.collectables.append(collectable_item)

