#!/usr/bin/env xpyv

#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program  
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
# 
#-------------------------------------------------------------------------------------------------   

from os import path, environ, system
from subprocess import PIPE, Popen
from shutil import copyfile
from tempfile import mktemp, mkdtemp

from interfaces import IXIVDiag
from collector.collectable.files import SimpleFileCollectable as FileItem
from collector.collectable.commands import SimpleCommandCollectable as CommandItem
from collector.collectable.directories import SimpleDirectoryCollectable as DirectoryItem

from win32evtlog import BackupEventLog, OpenEventLog, CloseEventLog
from win32api import GetComputerName
from xpyv import Wow64FsRedirection,project_root_path

from xiv_xsf.console import print_in_color

HOST_ATTACHMENT_DIR = path.abspath(path.join(project_root_path.find(False)))
DEFAULTDIR = path.join(environ['SystemRoot'], 'Temp')
CLUSTER_2008_EXE_PATH = path.join(environ['SystemRoot'], 'System32', 'cluster.exe')
CLUSTER_2003_LOG_PATH = path.join(environ['SystemRoot'], 'Cluster', 'cluster.log')
CLUSTER_2008_LOG_PATH = path.join(environ['SystemRoot'], 'System32', 'winevt', 'Logs', 'ClusterLog.etl')
CMD_PATH = path.join(environ['SystemRoot'], 'System32', 'cmd.exe')
DISKPART_PATH = path.join(environ['SystemRoot'], 'System32', 'diskpart.exe')
REG_PATH = path.join(environ['SystemRoot'], 'System32', 'reg.exe')
MSINFO32_PATH = path.join(environ['ProgramFiles'].strip(" (x86)"), 'Common files', 'Microsoft shared', 'MSinfo', 'msinfo32.exe')
MPIO_LIST_PATH = path.join(HOST_ATTACHMENT_DIR, 'drivers', 'mpio', 'mpiolist.exe')
XIV_DEVLIST_PATH = path.join(HOST_ATTACHMENT_DIR, 'bin', 'xiv_devlist.exe')
XIV_FC_ADMIN = path.join(HOST_ATTACHMENT_DIR, 'bin', 'xiv_fc_admin.exe')
XIV_ISCSI_ADMIN = path.join(HOST_ATTACHMENT_DIR, 'bin', 'xiv_iscsi_admin.exe')
MINIDUMP_PATH = path.join(environ['SystemRoot'], 'Minidump')
MEMORYDUMP_PATH = path.join(environ['SystemRoot'], 'MEMORY.DMP')
SETUPAPI_NT5_PATH = path.join(environ['SystemRoot'], 'setupapi.log')
SETUPAPI_NT6_PATH = path.join(environ['SystemRoot'], 'setupapi.log')
TEMP_PATH = path.join(environ['SystemRoot'], 'Temp')
INQUIRY_PATH = path.join(HOST_ATTACHMENT_DIR, 'lib', 'xiv_diag', 'scripts', 'inquiry.py')
WINDRIVERS_PATH = path.join(HOST_ATTACHMENT_DIR, 'lib', 'xiv_diag', 'scripts', 'windrivers.py')
MPATH_MGMT_AGENT_DATA_PATH = path.join(HOST_ATTACHMENT_DIR, 'data')

class XIVDiag(IXIVDiag):
    def __init__(self):
        self.get_computername()
        IXIVDiag.__init__(self, DEFAULTDIR)
        Wow64FsRedirection.disable()

    def get_computername(self):
        self.computername = GetComputerName()

    def add_collectables(self):
        self.add_msinfo32_collectables()
        self.add_eventlogs_collectables()
        self.add_clusterlog_collectables()
        self.add_mpiolist_collectables()
        self.add_diskpart_collectables()
        self.add_registry_collectables()
        self.add_services_collectables()
        self.add_kernel_dumps_collectables()
        self.add_setupapi_collectables()
        self.add_reg_hardware_subtree()
        self.add_xiv_devlist()
        self.add_xiv_admins()
        self.add_hostattachment_logs()
        self.add_xivdiag_scripts()
        self.add_xiv_mpath_mgmt_agent_data()

    def add_hostattachment_logs(self):
        (temp_file, temp_dir) = self.copy_hostattachment_logs_to_tempdir()
        self.add_collectable_item(FileItem("XIV Host Attachment Logs (1/2)", self.zip_file, temp_file))
        self.add_collectable_item(DirectoryItem("XIV Host Attachment Logs (2/2)", self.zip_file, temp_dir))
        
    def add_msinfo32_collectables(self):
        temp_file = mktemp(prefix='msinfo32_')
        temp_file = '%s.nfo' % temp_file
        self.add_msinfo32_command_collectable(temp_file)
        self.add_msinfo32_file_collectable(temp_file)

    def add_msinfo32_command_collectable(self, temp_file):
        self.add_collectable_item(CommandItem( \
            "System Information (1/2)", [MSINFO32_PATH, '/nfo', temp_file], timeout=180))

    def add_msinfo32_file_collectable(self, temp_file):
        self.add_collectable_item(FileItem("System Information (2/2)", self.zip_file, temp_file))

    def add_eventlogs_collectables(self):
        for eventlog in ('System', 'Application'):
            eventlog_tempfile = self.prepare_eventlog(eventlog)
            self.add_collectable_item(FileItem("%s Event Log" % eventlog, self.zip_file, eventlog_tempfile))

    def prepare_eventlog(self, eventlog):
        eventlog_tempfile = mktemp(prefix='%s_eventlog' % eventlog)
        try:
            eventlog_object = OpenEventLog(self.computername, eventlog)
            BackupEventLog(eventlog_object, eventlog_tempfile)
            CloseEventLog(eventlog_object)
        except:
            pass
        return eventlog_tempfile

    def add_clusterlog_collectables(self):
        self.add_collectable_item(FileItem("Cluster (2003) Log", self.zip_file, CLUSTER_2003_LOG_PATH)) 
        self.add_collectable_item(CommandItem("Cluster (2008) Log Generator", [CLUSTER_2008_EXE_PATH, 'log', '/generate']))
        for i in range(1,6):
            self.add_collectable_item(FileItem("Cluster (2008) Logs (%s/5)" % i, self.zip_file, '%s.00%s' % (CLUSTER_2008_LOG_PATH, i)))

    def add_mpiolist_collectables(self):
        for i in range(0,6):
            self.add_collectable_item(CommandItem("MultiPath I/O Information (%s/6)" % str(i+1), [MPIO_LIST_PATH, str(i)]))
    
    def add_diskpart_collectables(self):
        self.add_collectable_item(CommandItem('DISKPART: List Disk', [CMD_PATH, '/c echo LIST DISK | %s' % DISKPART_PATH]))
        self.add_collectable_item(CommandItem('DISKPART: List Volume', [CMD_PATH, '/c echo LIST VOLUME | %s' % DISKPART_PATH]))

    def add_registry_collectables(self):
        self.add_collectable_item(CommandItem('Installed HotFixes', \
            [REG_PATH, 'query', '''HKLM\SOFTWARE\Microsoft\Windows NT\CurrentVersion\HotFix''', '/s']))
        self.add_collectable_item(CommandItem('DSMXIV Configuration', \
            [REG_PATH, 'query', '''HKLM\SYSTEM\CurrentControlSet\Services\dsmxiv''', '/s']))

    def add_services_collectables(self):
        self.add_collectable_item(CommandItem('Services Information', ['sc', 'query']))
    
    def add_kernel_dumps_collectables(self):
        minidump_temp_file, memorydump_temp_file = self.copy_kernel_dumps_to_tempdir()
        self.add_collectable_item(FileItem('Windows Minidump', self.zip_file, minidump_temp_file))
        self.add_collectable_item(FileItem('Windows Memory Dump', self.zip_file, memorydump_temp_file))

    def add_setupapi_collectables(self):
        self.add_collectable_item(FileItem('Windows Setup API (1/2)', self.zip_file, SETUPAPI_NT5_PATH))
        self.add_collectable_item(FileItem('Windows Setup API (2/2)', self.zip_file, SETUPAPI_NT6_PATH))

    def add_reg_hardware_subtree(self):
        self.add_collectable_item(CommandItem('Hardware Registry Subtree', \
            [REG_PATH, 'query', '''HKLM\HARDWARE''', '/s']))

    def copy_kernel_dumps_to_tempdir(self):
        line = 'INFO: Copying Memory dumps to temporary directory... '.ljust(70, ' ')
        print_in_color(line, newline=False)
        output_tempfile = mktemp(prefix='output_')
        minidump_tempfile = mktemp(prefix='minidump_')
        memorydump_tempfile = mktemp(prefix='memorydump_')
        output = system('copy /y %s %s > %s' % (MINIDUMP_PATH, minidump_tempfile, output_tempfile))
        output = system('copy /y %s %s > %s' % (MEMORYDUMP_PATH, memorydump_tempfile, output_tempfile))
        print_in_color('DONE'.rjust(9, ' '), 'dark_green')
        return (minidump_tempfile, memorydump_tempfile)
    
    def copy_hostattachment_logs_to_tempdir(self):
        line = 'INFO: Copying Host Attachment logs to temporary directory... '.ljust(70,' ')
        print_in_color(line, newline=False)
        output_tempfile = mktemp(prefix='output_')
        logs_tempdir = mkdtemp(prefix='hak_logs_')
        output = system('copy /y %s\\host_attach*log* %s\\ > %s' % (path.join(environ['SystemRoot'],'Temp'),
                                                      logs_tempdir, output_tempfile))
        output = system('copy /y %s\\xiv_mpio_agent*log* %s\\ > %s' % (path.join(environ['SystemRoot'],'Temp'),
                                                      logs_tempdir, output_tempfile))
 
        print_in_color('DONE'.rjust(9, ' '), 'dark_green')
        return (output_tempfile, logs_tempdir)
        
    def add_xiv_devlist(self):
        self.add_collectable_item(CommandItem('xiv_devlist' , [XIV_DEVLIST_PATH, '-o', 'all']))

    def add_xivdiag_scripts(self):
        self.add_collectable_item(CommandItem('inquiry.py', ['xpyv', INQUIRY_PATH]))
        self.add_collectable_item(CommandItem('windrivers.py', ['xpyv', WINDRIVERS_PATH]))

    def add_xiv_admins(self):
        self.add_collectable_item(CommandItem('xiv_fc_admin -L', [XIV_FC_ADMIN, '-L']))
        self.add_collectable_item(CommandItem('xiv_fc_admin -V', [XIV_FC_ADMIN, '-V']))
        self.add_collectable_item(CommandItem('xiv_fc_admin -P', [XIV_FC_ADMIN, '-P']))
        self.add_collectable_item(CommandItem('xiv_iscsi_admin -L', [XIV_ISCSI_ADMIN, '-L']))
        self.add_collectable_item(CommandItem('xiv_iscsi_admin -V', [XIV_ISCSI_ADMIN, '-V']))
        self.add_collectable_item(CommandItem('xiv_iscsi_admin -P', [XIV_ISCSI_ADMIN, '-P']))

    def add_xiv_mpath_mgmt_agent_data(self):
        self.add_collectable_item(DirectoryItem("XIV Multipath I/O Agent Data", self.zip_file, MPATH_MGMT_AGENT_DATA_PATH))

